from __future__ import annotations

import subprocess
import sys
from textwrap import dedent
from typing import TYPE_CHECKING

import kuzu
import pytest

if TYPE_CHECKING:
    from pathlib import Path


def open_database_on_subprocess(tmp_path: Path, build_dir: Path) -> None:
    code = dedent(
        f"""
        import sys
        sys.path.append(r"{build_dir!s}")

        import kuzu
        db = kuzu.Database(r"{tmp_path!s}")
        print(r"{tmp_path!s}")
    """
    )
    result = subprocess.run([sys.executable, "-c", code], capture_output=True, text=True)
    if result.returncode != 0:
        raise RuntimeError(result.stderr)


def test_database_close(tmp_path: Path, build_dir: Path) -> None:
    db_path = tmp_path / "test_database_close.kuzu"
    db = kuzu.Database(database_path=db_path, read_only=False)
    assert not db.is_closed
    assert db._database is not None

    # Open the database on a subprocess. It should raise an exception.
    with pytest.raises(
        Exception,
        match=r"RuntimeError: IO exception: Could not set lock on file",
    ):
        open_database_on_subprocess(db_path, build_dir)

    db.close()
    assert db.is_closed
    assert db._database is None

    # Try to close the database again. It should not raise any exceptions.
    db.close()
    assert db.is_closed
    assert db._database is None

    # Open the database on a subprocess. It should not raise any exceptions.
    # TODO: Enable this test on Windows when the read-only mode is implemented.
    if sys.platform != "win32":
        open_database_on_subprocess(db_path, build_dir)


def test_database_context_manager(tmp_path: Path, build_dir: Path) -> None:
    db_path = tmp_path / "test_database_context_manager.kuzu"
    with kuzu.Database(database_path=db_path, read_only=False) as db:
        assert not db.is_closed
        assert db._database is not None

        # Open the database on a subprocess. It should raise an exception.
        # TODO: Enable this test on Windows when the read-only mode is implemented.
        if sys.platform != "win32":
            with pytest.raises(
                RuntimeError,
                match=r"RuntimeError: IO exception: Could not set lock on file",
            ):
                open_database_on_subprocess(db_path, build_dir)

    # Open the database on a subprocess. It should not raise any exceptions.
    # TODO: Enable this test on Windows when the read-only mode is implemented.
    if sys.platform != "win32":
        open_database_on_subprocess(db_path, build_dir)


def test_in_mem_database_memory_db_path() -> None:
    db = kuzu.Database(database_path=":memory:")
    assert not db.is_closed
    assert db._database is not None

    # Open the database on a subprocess. It should raise an exception.
    conn = kuzu.Connection(db)
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2


def test_in_mem_database_empty_db_path() -> None:
    db = kuzu.Database()
    assert not db.is_closed
    assert db._database is not None

    # Open the database on a subprocess. It should raise an exception.
    conn = kuzu.Connection(db)
    conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
    conn.execute("CREATE (:person {name: 'Alice', age: 30});")
    conn.execute("CREATE (:person {name: 'Bob', age: 40});")
    result = conn.execute("MATCH (p:person) RETURN p.*")
    assert result.get_num_tuples() == 2


def test_in_mem_database_no_db_path() -> None:
    with kuzu.Database(database_path="") as db:
        assert not db.is_closed
        assert db._database is not None

        # Open the database on a subprocess. It should raise an exception.
        conn = kuzu.Connection(db)
        conn.execute("CREATE NODE TABLE person(name STRING, age INT64, PRIMARY KEY(name));")
        conn.execute("CREATE (:person {name: 'Alice', age: 30});")
        conn.execute("CREATE (:person {name: 'Bob', age: 40});")
        with conn.execute("MATCH (p:person) RETURN p.*") as result:
            assert result.get_num_tuples() == 2


def test_database_auto_checkpoint_config(tmp_path: Path) -> None:
    with kuzu.Database(database_path=tmp_path, auto_checkpoint=False) as db:
        assert not db.is_closed
        assert db._database is not None

        conn = kuzu.Connection(db)
        with conn.execute("CALL current_setting('auto_checkpoint') RETURN *") as result:
            assert result.get_num_tuples() == 1
            assert result.get_next()[0] == "False"


def test_database_checkpoint_threshold_config(tmp_path: Path) -> None:
    with kuzu.Database(database_path=tmp_path, checkpoint_threshold=1234) as db:
        assert not db.is_closed
        assert db._database is not None

        conn = kuzu.Connection(db)
        with conn.execute("CALL current_setting('checkpoint_threshold') RETURN *") as result:
            assert result.get_num_tuples() == 1
            assert result.get_next()[0] == "1234"



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable

from . import _kuzu
from .prepared_statement import PreparedStatement
from .query_result import QueryResult

if TYPE_CHECKING:
    import sys
    from types import TracebackType

    from .database import Database
    from .types import Type

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


class Connection:
    """Connection to a database."""

    def __init__(self, database: Database, num_threads: int = 0):
        """
        Initialise kuzu database connection.

        Parameters
        ----------
        database : Database
            Database to connect to.

        num_threads : int
            Maximum number of threads to use for executing queries.

        """
        self._connection: Any = None  # (type: _kuzu.Connection from pybind11)
        self.database = database
        self.num_threads = num_threads
        self.is_closed = False
        self.init_connection()

    def __getstate__(self) -> dict[str, Any]:
        state = {
            "database": self.database,
            "num_threads": self.num_threads,
            "_connection": None,
        }
        return state

    def init_connection(self) -> None:
        """Establish a connection to the database, if not already initalised."""
        if self.is_closed:
            error_msg = "Connection is closed."
            raise RuntimeError(error_msg)
        self.database.init_database()
        if self._connection is None:
            self._connection = _kuzu.Connection(self.database._database, self.num_threads)  # type: ignore[union-attr]

    def set_max_threads_for_exec(self, num_threads: int) -> None:
        """
        Set the maximum number of threads for executing queries.

        Parameters
        ----------
        num_threads : int
            Maximum number of threads to use for executing queries.

        """
        self.init_connection()
        self._connection.set_max_threads_for_exec(num_threads)

    def close(self) -> None:
        """
        Close the connection.

        Note: Call to this method is optional. The connection will be closed
        automatically when the object goes out of scope.
        """
        if self._connection is not None:
            self._connection.close()
        self._connection = None
        self.is_closed = True

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        self.close()

    def execute(
        self,
        query: str | PreparedStatement,
        parameters: dict[str, Any] | None = None,
    ) -> QueryResult | list[QueryResult]:
        """
        Execute a query.

        Parameters
        ----------
        query : str | PreparedStatement
            A prepared statement or a query string.
            If a query string is given, a prepared statement will be created
            automatically.

        parameters : dict[str, Any]
            Parameters for the query.

        Returns
        -------
        QueryResult
            Query result.

        """
        if parameters is None:
            parameters = {}

        self.init_connection()
        if not isinstance(parameters, dict):
            msg = f"Parameters must be a dict; found {type(parameters)}."
            raise RuntimeError(msg)  # noqa: TRY004

        if len(parameters) == 0 and isinstance(query, str):
            _query_result = self._connection.query(query)
        else:
            prepared_statement = self.prepare(query) if isinstance(query, str) else query
            _query_result = self._connection.execute(prepared_statement._prepared_statement, parameters)
        if not _query_result.isSuccess():
            raise RuntimeError(_query_result.getErrorMessage())
        current_query_result = QueryResult(self, _query_result)
        if not _query_result.hasNextQueryResult():
            return current_query_result
        all_query_results = [current_query_result]
        while _query_result.hasNextQueryResult():
            _query_result = _query_result.getNextQueryResult()
            if not _query_result.isSuccess():
                raise RuntimeError(_query_result.getErrorMessage())
            all_query_results.append(QueryResult(self, _query_result))
        return all_query_results

    def prepare(self, query: str) -> PreparedStatement:
        """
        Create a prepared statement for a query.

        Parameters
        ----------
        query : str
            Query to prepare.

        Returns
        -------
        PreparedStatement
            Prepared statement.

        """
        return PreparedStatement(self, query)

    def _get_node_property_names(self, table_name: str) -> dict[str, Any]:
        LIST_START_SYMBOL = "["
        LIST_END_SYMBOL = "]"
        self.init_connection()
        query_result = self.execute(f"CALL table_info('{table_name}') RETURN *;")
        results = {}
        while query_result.has_next():
            row = query_result.get_next()
            prop_name = row[1]
            prop_type = row[2]
            is_primary_key = row[4] is True
            dimension = prop_type.count(LIST_START_SYMBOL)
            splitted = prop_type.split(LIST_START_SYMBOL)
            shape = []
            for s in splitted:
                if LIST_END_SYMBOL not in s:
                    continue
                s = s.split(LIST_END_SYMBOL)[0]
                if s != "":
                    shape.append(int(s))
            prop_type = splitted[0]
            results[prop_name] = {
                "type": prop_type,
                "dimension": dimension,
                "is_primary_key": is_primary_key,
            }
            if len(shape) > 0:
                results[prop_name]["shape"] = tuple(shape)
        return results

    def _get_node_table_names(self) -> list[Any]:
        results = []
        self.init_connection()
        query_result = self.execute("CALL show_tables() RETURN *;")
        while query_result.has_next():
            row = query_result.get_next()
            if row[2] == "NODE":
                results.append(row[1])
        return results

    def _get_rel_table_names(self) -> list[dict[str, Any]]:
        results = []
        self.init_connection()
        tables_result = self.execute("CALL show_tables() RETURN *;")
        while tables_result.has_next():
            row = tables_result.get_next()
            if row[2] == "REL":
                name = row[1]
                connections_result = self.execute(f"CALL show_connection({name!r}) RETURN *;")
                src_dst_row = connections_result.get_next()
                src_node = src_dst_row[0]
                dst_node = src_dst_row[1]
                results.append({"name": name, "src": src_node, "dst": dst_node})
        return results

    def set_query_timeout(self, timeout_in_ms: int) -> None:
        """
        Set the query timeout value in ms for executing queries.

        Parameters
        ----------
        timeout_in_ms : int
            query timeout value in ms for executing queries.

        """
        self.init_connection()
        self._connection.set_query_timeout(timeout_in_ms)

    def interrupt(self) -> None:
        """
        Interrupts execution of the current query.

        If there is no currently executing query, this function does nothing.
        """
        self._connection.interrupt()

    def create_function(
        self,
        name: str,
        udf: Callable[[...], Any],
        params_type: list[Type | str] | None = None,
        return_type: Type | str = "",
        *,
        default_null_handling: bool = True,
        catch_exceptions: bool = False,
    ) -> None:
        """
        Sets a User Defined Function (UDF) to use in cypher queries.

        Parameters
        ----------
        name: str
            name of function

        udf: Callable[[...], Any]
            function to be executed

        params_type: Optional[list[Type]]
            list of Type enums to describe the input parameters

        return_type: Optional[Type]
            a Type enum to describe the returned value

        default_null_handling: Optional[bool]
            if true, when any parameter is null, the resulting value will be null

        catch_exceptions: Optional[bool]
            if true, when an exception is thrown from python, the function output will be null
            Otherwise, the exception will be rethrown
        """
        if params_type is None:
            params_type = []
        parsed_params_type = [x if type(x) is str else x.value for x in params_type]
        if type(return_type) is not str:
            return_type = return_type.value

        self._connection.create_function(
            name=name,
            udf=udf,
            params_type=parsed_params_type,
            return_value=return_type,
            default_null=default_null_handling,
            catch_exceptions=catch_exceptions,
        )

    def remove_function(self, name: str) -> None:
        """
        Removes a User Defined Function (UDF).

        Parameters
        ----------
        name: str
            name of function to be removed.
        """
        self._connection.remove_function(name)



import os
import sys
import shutil
import logging
from hashlib import md5
from pathlib import Path

logging.basicConfig(level=logging.DEBUG)

HEADER_BASE_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), '../../src/include'))
MAIN_HEADER_PATH = os.path.realpath(
    os.path.join(HEADER_BASE_PATH, 'main'))
START_POINT = os.path.realpath(
    os.path.join(MAIN_HEADER_PATH, 'kuzu.h')
)
JSON_HEADER_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), '../../third_party/nlohmann_json/json_fwd.hpp'))
HEADER_TARGET_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), 'headers')
)

logging.debug('HEADER_BASE_PATH: %s', HEADER_BASE_PATH)
logging.debug('MAIN_HEADER_PATH: %s', MAIN_HEADER_PATH)
logging.debug('START_POINT: %s', START_POINT)
logging.debug('JSON_HEADER_PATH: %s', JSON_HEADER_PATH)


def collect_header_file_path_recurse(start_point):
    global processed_header_files
    global header_map
    if start_point in processed_header_files:
        return []
    curr_header_files = []
    with open(start_point, 'r') as f:
        for line in f.readlines():
            if not line.startswith('#include "'):
                continue
            header_path = os.path.normpath(line.split('"')[1])
            header_real_path = None
            # Special case for json_fwd.hpp
            if header_path == 'json_fwd.hpp':
                header_real_path = JSON_HEADER_PATH
                logging.debug('Found header: %s at %s',
                              header_path, header_real_path)
            else:
                # Check if the header is in the current directory
                start_point_dir = os.path.dirname(start_point)
                header_real_path = os.path.join(
                    start_point_dir, header_path)
                if os.path.exists(header_real_path):
                    logging.debug('Found header: %s at %s',
                                  header_path, header_real_path)
                else:
                    # Check if the header is in the include directory
                    header_real_path = os.path.join(
                        HEADER_BASE_PATH, header_path)
                    if os.path.exists(header_real_path):
                        logging.debug('Found header: %s at %s',
                                      header_path, header_real_path)
            if header_real_path is None:
                logging.error('Could not find header: %s', header_path)
                sys.exit(1)
            curr_header_files.append(header_real_path)
            if start_point not in header_map:
                header_map[start_point] = {}
            header_map[start_point][header_path] = header_real_path
    for header_file in curr_header_files:
        curr_header_files += collect_header_file_path_recurse(header_file)

    processed_header_files.add(start_point)
    return curr_header_files


def collect_header_file_paths():
    global processed_header_files
    global header_map
    processed_header_files = set()
    header_map = {}
    collect_header_file_path_recurse(START_POINT)


def copy_header(header_real_path):
    global copied_headers
    if header_real_path in copied_headers:
        return copied_headers[header_real_path]
    header_name = os.path.basename(header_real_path)
    # Rename the header if it is already copied
    if os.path.exists(os.path.join(HEADER_TARGET_PATH, header_name)):
        header_name = md5(header_real_path.encode()).hexdigest() + '.h'
    target_path = os.path.join(HEADER_TARGET_PATH, header_name)
    shutil.copyfile(header_real_path, target_path)
    copied_headers[header_real_path] = header_name
    logging.debug('Copied header: %s to %s', header_real_path, target_path)
    return header_name


def copy_headers():
    global header_map
    global copied_headers
    copied_headers = {}
    if os.path.exists(HEADER_TARGET_PATH):
        shutil.rmtree(HEADER_TARGET_PATH, ignore_errors=True)
    os.makedirs(HEADER_TARGET_PATH)
    for src_header in header_map:
        src_header_copied_name = copy_header(src_header)
        src_header_copied_path = os.path.join(
            HEADER_TARGET_PATH, src_header_copied_name)
        file = Path(src_header_copied_path)
        for original_header_path in header_map[src_header]:
            header_real_path = header_map[src_header][original_header_path]
            header_name = copy_header(header_real_path)
            file.write_text(file.read_text().replace(
                original_header_path, header_name))


if __name__ == '__main__':
    logging.info('Collecting header files...')
    collect_header_file_paths()
    logging.info('Copying header files...')
    copy_headers()
    logging.info('Done!')



# Step 01 - download RFC7932.
#
# RFC is the ultimate source for brotli format and constants, including
# static dictionary.

import urllib2

response = urllib2.urlopen("https://tools.ietf.org/rfc/rfc7932.txt")

text = response.read()
path = "rfc7932.txt"

with open(path, "w") as rfc:
  rfc.write(text)

print("Downloaded and saved " + str(len(text)) + " bytes to " + path)



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

import pytest
from type_aliases import ConnDB

# required by python-lint
if TYPE_CHECKING:
    from pathlib import Path
import kuzu


def test_struct_param_access(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite
    batch = [
        {
            "id": "1",
            "name": "Christopher",
            "age": "61",
            "net_worth": "21561030.8",
            "email": "wilcox@hotmail.com",
            "address": "588 Campbell Well Suite 335, Lawrencemouth, VI 07241",
            "phone": "358-202-1821x630",
            "comments": "Stage modern card send point future serve. Hot month might pass dinner chance.\nPerformance become own spring from analysis commercial. Draw effort structure many drug first.",
        },
        {
            "id": "2",
            "name": "Amanda",
            "age": "42",
            "net_worth": "31344480.2",
            "email": "cook@hotmail.com",
            "address": "6293 Bright Centers, Chenton, MO 36829",
            "phone": "001-972-387-3913x966",
            "comments": "None agent some if skill. Often onto dog wish. Listen live hair garden contain worry time. Economic or institution statement energy take sit.",
        },
    ]
    conn.execute(
        """
        UNWIND $batch AS p
        RETURN p.name,
            p.age,
            p.net_worth,
            p.email,
            p.address,
            p.phone,
            p.comments
        """,
        parameters={"batch": batch},
    )


def test_array_binding(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite
    conn.execute("CREATE NODE TABLE node(id STRING, embedding DOUBLE[3], PRIMARY KEY(id))")
    conn.execute("CREATE (d:node {id: 'test', embedding: $emb})", {"emb": [3, 5, 2]})
    result = conn.execute(
        """
        MATCH (d:node)
        RETURN d.id, array_cosine_similarity(d.embedding, $emb)
        """,
        {"emb": [4.3, 5.2, 6.7]},
    )
    assert result.get_next() == ["test", pytest.approx(0.8922316795174099)]
    # TODO(maxwell): fixme. The following case should be executed successfully.
    # with pytest.raises(RuntimeError) as err:
    #     conn.execute(
    #         """
    #         MATCH (d:node)
    #         RETURN d.id, array_cosine_similarity($emb1, $emb)
    #         """, {"emb": [4.3, 5.2, 6.7], "emb1": [2.2, 3.3, 5.5]}
    #     )
    # assert str(err.value) == "Binder exception: Left and right type are both ANY, which is not currently supported."


def test_bool_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute(
        "MATCH (a:person) WHERE a.isStudent = $1 AND a.isWorker = $k RETURN COUNT(*)", {"1": False, "k": False}
    )
    assert result.has_next()
    assert result.get_next() == [1]
    assert not result.has_next()
    result.close()


def test_int_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.age < $AGE RETURN COUNT(*)", {"AGE": 1})
    assert result.has_next()
    assert result.get_next() == [0]
    assert not result.has_next()
    result.close()


def test_double_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.eyeSight = $E RETURN COUNT(*)", {"E": 5.0})
    assert result.has_next()
    assert result.get_next() == [2]
    assert not result.has_next()
    result.close()


def test_str_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN concat(a.fName, $S);", {"S": "HH"})
    assert result.has_next()
    assert result.get_next() == ["AliceHH"]
    assert not result.has_next()
    result.close()


def test_date_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("MATCH (a:person) WHERE a.birthdate = $1 RETURN COUNT(*);", {"1": datetime.date(1900, 1, 1)})
    assert result.has_next()
    assert result.get_next() == [2]
    assert not result.has_next()
    result.close()


def test_timestamp_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute(
        "MATCH (a:person) WHERE a.registerTime = $1 RETURN COUNT(*);",
        {"1": datetime.datetime(2011, 8, 20, 11, 25, 30)},
    )
    assert result.has_next()
    assert result.get_next() == [1]
    assert not result.has_next()
    result.close()


def test_int64_list_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("MATCH (a:person {workedHours: $1}) RETURN COUNT(*);", {"1": [3, 4, 5, 6, 7]})
    assert result.has_next()
    assert result.get_next() == [1]
    assert not result.has_next()
    result.close()

def test_empty_list_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("RETURN list_contains($list, $item)", {"list": [], "item": 5})
    assert result.has_next()
    assert result.get_next() == [False]
    assert not result.has_next()
    result.close()

def test_int64_list_list_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute(
        "MATCH (a:person) WHERE a.courseScoresPerTerm = $1 OR a.courseScoresPerTerm = $2 RETURN COUNT(*);",
        {"1": [[8, 10]], "2": [[7, 4], [8, 8], [9]]},
    )
    assert result.has_next()
    assert result.get_next() == [2]
    assert not result.has_next()
    result.close()


def test_string_list_param(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    result = conn.execute("MATCH (a:person {usedNames: $1}) RETURN COUNT(*);", {"1": ["Carmen", "Fred"]})
    assert result.has_next()
    assert result.get_next() == [1]
    assert not result.has_next()
    result.close()


def test_map_param(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute(
        "CREATE NODE TABLE tab(id int64, mp MAP(double, int64), mp2 MAP(int64, double), mp3 MAP(string, string), mp4 MAP(string, string)[], primary key(id))"
    )
    result = conn.execute(
        "MERGE (t:tab {id: 0, mp: $1, mp2: $2, mp3: $3, mp4: $4}) RETURN t.*",
        {
            "1": {"key": [1.0, 2, 2.2], "value": [5, 3, -1]},
            "2": {"key": [5, 4, 0], "value": [-0.5, 0, 2.2]},
            "3": {"key": ["a", "b", "c"], "value": [1, "2", "3"]},
            "4": [{"key": ["a"], "value": ["b"]}],
        },
    )
    assert result.has_next()
    assert result.get_next() == [
        0,
        {1.0: 5, 2.0: 3, 2.2: -1},
        {5: -0.5, 4: -0.0, 0: 2.2},
        {"a": "1", "b": "2", "c": "3"},
        [{"a": "b"}],
    ]
    assert not result.has_next()
    result.close()


def test_general_list_param(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute(
        "CREATE NODE TABLE tab(id int64, lst1 BOOL[], lst2 DOUBLE[], lst3 TIMESTAMP[], lst4 DATE[], lst5 INTERVAL[], lst6 STRING[], PRIMARY KEY(id))"
    )
    lst1 = [True, False]
    lst2 = [1.0, 2.0]
    lst3 = [datetime.datetime(2019, 11, 12, 11, 25, 30), datetime.datetime(1987, 2, 15, 3, 0, 2)]
    lst4 = [datetime.date(2019, 11, 12), datetime.date(1987, 2, 15)]
    lst5 = [lst3[0] - lst3[1]]
    lst6 = [1, "2", "3"]
    lst6ToString = ["1", "2", "3"]
    result = conn.execute(
        "MERGE (t:tab {id: 0, lst1: $1, lst2: $2, lst3: $3, lst4: $4, lst5: $5, lst6: $6}) RETURN t.*",
        {"1": lst1, "2": lst2, "3": lst3, "4": lst4, "5": lst5, "6": lst6},
    )
    assert result.has_next()
    assert result.get_next() == [0, lst1, lst2, lst3, lst4, lst5, lst6ToString]
    assert not result.has_next()
    result.close()


def test_null_resolution(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute(
        "CREATE NODE TABLE tab(id SERIAL, lst1 INT64[], mp1 MAP(STRING, STRING), "
        "nest MAP(STRING, MAP(STRING, INT64))[], PRIMARY KEY(id))"
    )
    lst1 = [1, 2, 3, None]
    mp1 = {"key": ["a", "b", "c", "o"], "value": ["x", "y", "z", None]}
    nest = [
        {"key": ["2"], "value": [{"key": ["foo", "bar"], "value": [1, 2]}]},
        {"key": ["1"], "value": [{"key": [], "value": []}]},
    ]
    result = conn.execute("MERGE (t:tab {lst1: $1, mp1: $2, nest: $3}) RETURN t.*", {"1": lst1, "2": mp1, "3": nest})
    assert result.has_next()
    assert result.get_next() == [
        0,
        lst1,
        {"a": "x", "b": "y", "c": "z", "o": None},
        [{"2": {"foo": 1, "bar": 2}}, {"1": {}}],
    ]
    assert not result.has_next()
    result.close()


def test_param_error1(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    with pytest.raises(RuntimeError, match="Parameter name must be of type string but got <class 'int'>"):
        conn.execute("MATCH (a:person) WHERE a.registerTime = $1 RETURN COUNT(*);", {1: 1})


def test_param_error2(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    with pytest.raises(RuntimeError, match="Parameters must be a dict"):
        conn.execute("MATCH (a:person) WHERE a.registerTime = $1 RETURN COUNT(*);", ["asd"])


def test_param_error3(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    with pytest.raises(RuntimeError, match="Parameters must be a dict"):
        conn.execute("MATCH (a:person) WHERE a.registerTime = $1 RETURN COUNT(*);", [("asd", 1, 1)])


def test_param(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    conn.execute("CREATE NODE TABLE NodeOne(id INT64, name STRING, PRIMARY KEY(id));")
    conn.execute("CREATE NODE TABLE NodeTwo(id INT64, name STRING, PRIMARY KEY(id));")
    conn.execute("CREATE Rel TABLE RelA(from NodeOne to NodeOne);")
    conn.execute("CREATE Rel TABLE RelB(from NodeTwo to NodeOne, id int64, name String);")
    conn.execute('CREATE (t: NodeOne {id:1, name: "Alice"});')
    conn.execute('CREATE (t: NodeOne {id:2, name: "Jack"});')
    conn.execute('CREATE (t: NodeTwo {id:3, name: "Bob"});')
    result = conn.execute(
        "MATCH (a:NodeOne { id: $a_id }),"
        "(b:NodeTwo { id: $b_id }),"
        "(c: NodeOne{ id: $c_id } )"
        " MERGE"
        " (a)-[:RelA]->(c),"
        " (b)-[r:RelB { id: 2, name: $my_param }]->(c)"
        " return r.*;",
        {"a_id": 1, "b_id": 3, "c_id": 2, "my_param": None},
    )
    assert result.has_next()
    assert result.get_next() == [2, None]
    result.close()


def test_param_error4(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    with pytest.raises(
        RuntimeError,
        match="Runtime exception: Cannot convert Python object to Kuzu value : INT8  is incompatible with TIMESTAMP",
    ):
        conn.execute(
            "MATCH (a:person {workedHours: $1}) RETURN COUNT(*);", {"1": [1, 2, datetime.datetime(2023, 3, 25)]}
        )


def test_dict_conversion(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    # Interpret as MAP.
    result = conn.execute("RETURN $st", {"st": {"key": [1, 2, 3], "value": [3, 7, 98]}})
    assert result.get_next() == [{1: 3, 2: 7, 3: 98}]
    # Interpret as STRUCT since the first field name is not "key".
    result = conn.execute("RETURN $st", {"st": {"key1": [1, 2, 3], "value": [3, 7, 98]}})
    assert result.get_next() == [{"key1": [1, 2, 3], "value": [3, 7, 98]}]
    # Interpret as STRUCT since the number of elements in key and value doesn't match.
    result = conn.execute("RETURN $st", {"st": {"key": [1, 2], "value": [3, 7, 98, 4]}})
    assert result.get_next() == [{"key": [1, 2], "value": [3, 7, 98, 4]}]



import datetime
import re
from pathlib import Path
from uuid import UUID

import numpy as np
import pandas as pd
import pytest

try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo  # type: ignore[no-redef]

import kuzu


def validate_scan_pandas_results(results: kuzu.QueryResult) -> None:
    assert results.get_next() == [
        True,
        1,
        10,
        100,
        1000,
        -1,
        -10,
        -100,
        -1000,
        -0.5199999809265137,
        5132.12321,
        datetime.datetime(1996, 4, 1, 12, 0, 11, 500001),
        datetime.datetime(1996, 4, 1, 12, 0, 11, 500001, ZoneInfo("US/Eastern")),
        datetime.datetime(1996, 4, 1, 12, 0, 11, 500001),
        datetime.datetime(1996, 4, 1, 12, 0, 11, 500000),
        datetime.datetime(1996, 4, 1, 12, 0, 11),
        datetime.timedelta(microseconds=500),
        None,
        [],
        528,
        3.562,
        ["Alice", None],
        datetime.date(1996, 2, 15),
        "12331",
        UUID("d5a8ed71-6fc4-4cb3-acbc-2f5b73fd14bc"),
    ]
    assert results.get_next() == [
        False,
        2,
        20,
        200,
        2000,
        -2,
        -20,
        -200,
        -2000,
        None,
        24.222,
        datetime.datetime(1981, 11, 13, 22, 2, 52, 2),
        datetime.datetime(1981, 11, 13, 22, 2, 52, 2, ZoneInfo("US/Eastern")),
        datetime.datetime(1981, 11, 13, 22, 2, 52, 2),
        datetime.datetime(1981, 11, 13, 22, 2, 52),
        datetime.datetime(1981, 11, 13, 22, 2, 52),
        datetime.timedelta(seconds=1),
        "Ascii only",
        [40, 20, 10],
        -9999,
        4.213,
        [],
        datetime.date(2013, 2, 22),
        "test string",
        UUID("9a2fc988-5c5d-4217-af9e-220aef5ce7b8"),
    ]
    assert results.get_next() == [
        None,
        3,
        30,
        300,
        3000,
        -3,
        -30,
        -300,
        -3000,
        -3.299999952316284,
        None,
        datetime.datetime(1972, 12, 21, 12, 5, 44, 500003),
        datetime.datetime(1972, 12, 21, 12, 5, 44, 500003, ZoneInfo("US/Eastern")),
        datetime.datetime(1972, 12, 21, 12, 5, 44, 500003),
        datetime.datetime(1972, 12, 21, 12, 5, 44, 500000),
        datetime.datetime(1972, 12, 21, 12, 5, 44),
        datetime.timedelta(seconds=2, milliseconds=500),
        "ñ中国字",
        [30, None],
        None,
        None,
        None,
        datetime.date(2055, 1, 14),
        "5.623",
        UUID("166055ee-a481-4e67-a4fc-98682d3a3e20"),
    ]
    assert results.get_next() == [
        False,
        4,
        40,
        400,
        4000,
        -4,
        -40,
        -400,
        -4000,
        4.400000095367432,
        4.444,
        datetime.datetime(2008, 1, 11, 22, 10, 3, 4),
        datetime.datetime(2008, 1, 11, 22, 10, 3, 4, ZoneInfo("US/Eastern")),
        datetime.datetime(2008, 1, 11, 22, 10, 3, 4),
        datetime.datetime(2008, 1, 11, 22, 10, 3),
        datetime.datetime(2008, 1, 11, 22, 10, 3),
        datetime.timedelta(seconds=3, milliseconds=22),
        "😂",
        None,
        56677,
        67.13,
        ["Dan, Ella", "George"],
        datetime.date(2018, 3, 17),
        None,
        UUID("d5a8ed71-6fc4-4cb3-acbc-2f5b73fd14bc"),
    ]


def test_scan_pandas(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    data = {
        "BOOL": [True, False, None, False],
        "UINT8": np.array([1, 2, 3, 4], dtype=np.uint8),
        "UINT16": np.array([10, 20, 30, 40], dtype=np.uint16),
        "UINT32": np.array([100, 200, 300, 400], dtype=np.uint32),
        "UINT64": np.array([1000, 2000, 3000, 4000], dtype=np.uint64),
        "INT8": np.array([-1, -2, -3, -4], dtype=np.int8),
        "INT16": np.array([-10, -20, -30, -40], dtype=np.int16),
        "INT32": np.array([-100, -200, -300, -400], dtype=np.int32),
        "INT64": np.array([-1000, -2000, -3000, -4000], dtype=np.int64),
        "FLOAT_32": np.array(
            [-0.5199999809265137, float("nan"), -3.299999952316284, 4.400000095367432],
            dtype=np.float32,
        ),
        "FLOAT_64": np.array([5132.12321, 24.222, float("nan"), 4.444], dtype=np.float64),
        "datetime_microseconds": np.array([
            np.datetime64("1996-04-01T12:00:11.500001000"),
            np.datetime64("1981-11-13T22:02:52.000002000"),
            np.datetime64("1972-12-21T12:05:44.500003000"),
            np.datetime64("2008-01-11T22:10:03.000004000"),
        ]).astype("datetime64[us]"),
        "datetime_microseconds_tz": np.array([
            np.datetime64("1996-04-01T12:00:11.500001000"),
            np.datetime64("1981-11-13T22:02:52.000002000"),
            np.datetime64("1972-12-21T12:05:44.500003000"),
            np.datetime64("2008-01-11T22:10:03.000004000"),
        ]).astype("datetime64[us]"),
        "datetime_nanoseconds": np.array([
            np.datetime64("1996-04-01T12:00:11.500001"),
            np.datetime64("1981-11-13T22:02:52.000002"),
            np.datetime64("1972-12-21T12:05:44.500003"),
            np.datetime64("2008-01-11T22:10:03.000004"),
        ]).astype("datetime64[ns]"),
        "datetime_milliseconds": np.array([
            np.datetime64("1996-04-01T12:00:11.500001"),
            np.datetime64("1981-11-13T22:02:52.000002"),
            np.datetime64("1972-12-21T12:05:44.500003"),
            np.datetime64("2008-01-11T22:10:03.000004"),
        ]).astype("datetime64[ms]"),
        "datetime_seconds": np.array([
            np.datetime64("1996-04-01T12:00:11"),
            np.datetime64("1981-11-13T22:02:52"),
            np.datetime64("1972-12-21T12:05:44"),
            np.datetime64("2008-01-11T22:10:03"),
        ]).astype("datetime64[s]"),
        "timedelta_nanoseconds": [
            np.timedelta64(500000, "ns"),
            np.timedelta64(1000000000, "ns"),
            np.timedelta64(2500000000, "ns"),
            np.timedelta64(3022000000, "ns"),
        ],
        "name": [None, "Ascii only", "ñ中国字", "😂"],
        "worked_hours": [[], [40, 20, 10], [30, None], None],
        "int_object": np.array([528, -9999, None, 56677], dtype=object),
        "float_object": np.array([3.562, 4.213, None, 67.13], dtype=object),
        "used_names": np.array([["Alice", None], [], None, ["Dan, Ella", "George"]], dtype=object),
        "past_date": np.array(
            [
                datetime.date(1996, 2, 15),
                datetime.date(2013, 2, 22),
                datetime.date(2055, 1, 14),
                datetime.date(2018, 3, 17),
            ],
            dtype=object,
        ),
        "mixed_type": np.array([12331, "test string", 5.623, None], dtype="object"),
        "uuid_type": [
            UUID("d5a8ed71-6fc4-4cb3-acbc-2f5b73fd14bc"),
            UUID("9a2fc988-5c5d-4217-af9e-220aef5ce7b8"),
            UUID("166055ee-a481-4e67-a4fc-98682d3a3e20"),
            UUID("d5a8ed71-6fc4-4cb3-acbc-2f5b73fd14bc"),
        ],
    }
    df = pd.DataFrame(data)
    df["datetime_microseconds_tz"] = df["datetime_microseconds_tz"].dt.tz_localize("US/Eastern")
    results = conn.execute("LOAD FROM df RETURN *")
    validate_scan_pandas_results(results)


def test_scan_pandas_timestamp(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    ts = np.array(
        [
            datetime.datetime(1996, 2, 15, hour=12, minute=22, second=54),
            datetime.datetime(2011, 3, 11, minute=11, hour=5),
            None,
            datetime.datetime(2033, 2, 11, microsecond=55),
        ],
        dtype="object",
    )
    df = pd.DataFrame({"timestamp": ts})
    # Pandas automatically converts the column from object to timestamp, so we need to manually cast back to object.
    df = df.astype({"timestamp": "object"}, copy=False)
    results = conn.execute("LOAD FROM df RETURN *")
    assert results.get_next() == [datetime.datetime(1996, 2, 15, hour=12, minute=22, second=54)]
    assert results.get_next() == [datetime.datetime(2011, 3, 11, minute=11, hour=5)]
    assert results.get_next() == [None]
    assert results.get_next() == [datetime.datetime(2033, 2, 11, microsecond=55)]


def test_replace_failure(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)

    with pytest.raises(RuntimeError, match=re.escape("Binder exception: Variable x is not in scope.")):
        conn.execute("LOAD FROM x RETURN *;")

    with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Binder exception: Cannot match a built-in function for given function "
                "READ_PANDAS(STRING). Supported inputs are\n(POINTER)\n"
            ),
    ):
        conn.execute("CALL READ_PANDAS('df213') WHERE id > 20 RETURN id + 5, weight")


def test_int64_overflow(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    overflowpd = pd.DataFrame({"id": [4, 2 ** 125]})
    with pytest.raises(
            RuntimeError,
            match=re.escape(
                "Conversion exception: Failed to cast value: "
                "Python value '42535295865117307932921825928971026432' to INT64"
            ),
    ):
        conn.execute("LOAD FROM overflowpd RETURN *;")


def test_scan_pandas_with_filter(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    data = {
        "id": np.array([22, 3, 100], dtype=np.uint8),
        "weight": np.array([23.2, 31.7, 42.9], dtype=np.float64),
        "name": ["ñ", "日本字", "😊"],
    }
    df = pd.DataFrame(data)
    # Dummy query to ensure the READ_PANDAS function is persistent after a write transaction.
    conn.execute("CREATE NODE TABLE PERSON1(ID INT64, PRIMARY KEY(ID))")
    results = conn.execute("LOAD FROM df WHERE id > 20 RETURN id + 5, weight, name")
    assert results.get_next() == [27, 23.2, "ñ"]
    assert results.get_next() == [105, 42.9, "😊"]


def test_large_pd(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    num_rows = 40000
    odd_numbers = [2 * i + 1 for i in range(num_rows)]
    even_numbers = [2 * i for i in range(num_rows)]
    df = pd.DataFrame({
        "odd": np.array(odd_numbers, dtype=np.int64),
        "even": np.array(even_numbers, dtype=np.int64),
    })
    result = conn.execute("LOAD FROM df RETURN *").get_as_df()
    assert result["odd"].to_list() == odd_numbers
    assert result["even"].to_list() == even_numbers


def test_pandas_scan_demo(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)

    conn.execute("CREATE NODE TABLE student (ID int64, height int32, PRIMARY KEY(ID))")
    conn.execute("CREATE (s:student {ID: 0, height: 70})")
    conn.execute("CREATE (s:student {ID: 2, height: 64})")
    conn.execute("CREATE (s:student {ID: 4, height: 67})")
    conn.execute("CREATE (s:student {ID: 5, height: 64})")

    id = np.array([0, 2, 3, 5, 7, 11, 13], dtype=np.int64)
    age = np.array([42, 23, 33, 57, 67, 39, 11], dtype=np.uint16)
    height_in_cm = np.array([167, 172, 183, 199, 149, 154, 165], dtype=np.uint32)
    is_student = np.array([False, True, False, False, False, False, True], dtype=bool)
    person = pd.DataFrame({"id": id, "age": age, "height": height_in_cm, "is_student": is_student})

    result = conn.execute(
        "LOAD FROM person with avg(height / 2.54) as height_in_inch MATCH (s:student) WHERE s.height > "
        "height_in_inch RETURN s"
    ).get_as_df()
    assert len(result) == 2
    assert result["s"][0] == {
        "ID": 0,
        "_id": {"offset": 0, "table": 0},
        "_label": "student",
        "height": 70,
    }
    assert result["s"][1] == {
        "ID": 4,
        "_id": {"offset": 2, "table": 0},
        "_label": "student",
        "height": 67,
    }

    conn.execute("CREATE NODE TABLE person(ID INT64, age UINT16, height UINT32, is_student BOOLean, PRIMARY KEY(ID))")
    conn.execute("LOAD FROM person CREATE (p:person {ID: id, age: age, height: height, is_student: is_student})")
    result = conn.execute("MATCH (p:person) return p.*").get_as_df()
    assert np.all(result["p.ID"].to_list() == id)
    assert np.all(result["p.age"].to_list() == age)
    assert np.all(result["p.height"].to_list() == height_in_cm)
    assert np.all(result["p.is_student"].to_list() == is_student)


def test_scan_pandas_copy_subquery(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    data = {"id": np.array([22, 3, 100], dtype=np.int64), "name": ["A", "B", "C"]}
    df = pd.DataFrame(data)
    conn.execute("CREATE NODE TABLE person(ID INT64, NAME STRING, PRIMARY KEY(ID))")
    conn.execute("COPY person FROM (LOAD FROM df RETURN *)")
    result = conn.execute("MATCH (p:person) RETURN p.*").get_as_df()
    assert result["p.ID"].to_list() == [22, 3, 100]
    assert result["p.NAME"].to_list() == ["A", "B", "C"]


def test_scan_all_null(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    data = {"id": np.array([None, None, None], dtype=object)}
    df = pd.DataFrame(data)
    result = conn.execute("LOAD FROM df RETURN *")
    assert result.get_next() == [None]
    assert result.get_next() == [None]
    assert result.get_next() == [None]


def test_copy_from_scan_pandas_result(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age INT64, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM (LOAD FROM df WHERE age < 30 RETURN *);")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == ["Noura", 25]
    assert result.has_next() is False


def test_scan_from_py_arrow_pandas(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]}).convert_dtypes(
        dtype_backend="pyarrow"
    )
    result = conn.execute("LOAD FROM df RETURN *;")
    assert result.get_next() == ["Adam", 30]
    assert result.get_next() == ["Karissa", 40]
    assert result.get_next() == ["Zhang", 50]
    assert result.get_next() == ["Noura", 25]
    assert result.has_next() is False


def test_scan_long_utf8_string(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    data = {"name": ["很长的一段中文", "短", "非常长的中文"]}
    df = pd.DataFrame(data)
    result = conn.execute("LOAD FROM df WHERE name = '非常长的中文' RETURN count(*);")
    assert result.get_next() == [1]


def test_copy_from_pandas_object(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age STRING, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM df;")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == ["Adam", "30"]
    assert result.get_next() == ["Karissa", "40"]
    assert result.get_next() == ["Zhang", "50"]
    assert result.get_next() == ["Noura", "25"]
    assert result.has_next() is False
    df = pd.DataFrame({"f": ["Adam", "Karissa"], "t": ["Zhang", "Zhang"]})
    conn.execute("CREATE REL TABLE Knows(FROM Person TO Person);")
    conn.execute("COPY Knows FROM df")
    result = conn.execute("match (p:Person)-[]->(:Person {name: 'Zhang'}) return p.*")
    assert result.get_next() == ["Adam", "30"]
    assert result.get_next() == ["Karissa", "40"]
    assert result.has_next() is False


def test_copy_from_pandas_object_skip(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age STRING, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM df(SKIP=2);")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == ["Zhang", "50"]
    assert result.get_next() == ["Noura", "25"]
    assert result.has_next() is False
    df = pd.DataFrame({"f": ["Adam", "Noura"], "t": ["Zhang", "Zhang"]})
    conn.execute("CREATE REL TABLE Knows(FROM Person TO Person);")
    conn.execute("COPY Knows FROM df(SKIP=1)")
    result = conn.execute("match (p:Person)-[]->(:Person {name: 'Zhang'}) return p.*")
    assert result.get_next() == ["Noura", "25"]
    assert result.has_next() is False


def test_copy_from_pandas_object_limit(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age STRING, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM df(LIMIT=2);")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == ["Adam", "30"]
    assert result.get_next() == ["Karissa", "40"]
    assert result.has_next() is False
    df = pd.DataFrame({"f": ["Adam", "Zhang"], "t": ["Karissa", "Karissa"]})
    conn.execute("CREATE REL TABLE Knows(FROM Person TO Person);")
    conn.execute("COPY Knows FROM df(LIMIT=1)")
    result = conn.execute("match (p:Person)-[]->(:Person {name: 'Karissa'}) return p.*")
    assert result.get_next() == ["Adam", "30"]
    assert result.has_next() is False


def test_copy_from_pandas_object_skip_and_limit(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age STRING, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM df(SKIP=1, LIMIT=2);")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == ["Karissa", "40"]
    assert result.get_next() == ["Zhang", "50"]
    assert result.has_next() is False


def test_copy_from_pandas_object_skip_bounds_check(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age STRING, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM df(SKIP=10);")
    result = conn.execute("match (p:Person) return p.*")
    assert result.has_next() is False


def test_copy_from_pandas_object_limit_bounds_check(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"name": ["Adam", "Karissa", "Zhang", "Noura"], "age": [30, 40, 50, 25]})
    conn.execute("CREATE NODE TABLE Person(name STRING, age STRING, PRIMARY KEY (name));")
    conn.execute("COPY Person FROM df(LIMIT=10);")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == ["Adam", "30"]
    assert result.get_next() == ["Karissa", "40"]
    assert result.get_next() == ["Zhang", "50"]
    assert result.get_next() == ["Noura", "25"]
    assert result.has_next() is False


def test_copy_from_pandas_date(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"id": [1, 2], "date": [pd.Timestamp("2024-01-03"), pd.Timestamp("2023-10-10")]})
    conn.execute("CREATE NODE TABLE Person(id INT16, d TIMESTAMP, PRIMARY KEY (id));")
    conn.execute("COPY Person FROM df;")
    result = conn.execute("match (p:Person) return p.*")
    assert result.get_next() == [1, datetime.datetime(2024, 1, 3)]
    assert result.get_next() == [2, datetime.datetime(2023, 10, 10)]
    assert result.has_next() is False


def test_scan_string_to_nested(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({
        "id": ["1"],
        "lstcol": ["[1,2,3]"],
        "mapcol": ["{'a'=1,'b'=2}"],
        "structcol": ["{a:1,b:2}"],
        "lstlstcol": ["[[],[1,2,3],[4,5,6]]"],
    })
    conn.execute(
        "CREATE NODE TABLE tab(id INT64, lstcol INT64[], mapcol MAP(STRING, INT64), structcol STRUCT(a INT64, b INT64), lstlstcol INT64[][], PRIMARY KEY(id))"
    )
    conn.execute("COPY tab from df")
    result = conn.execute("match (t:tab) return t.*")
    assert result.get_next() == [
        1,
        [1, 2, 3],
        {"'a'": 1, "'b'": 2},
        {"a": 1, "b": 2},
        [[], [1, 2, 3], [4, 5, 6]],
    ]
    assert not result.has_next()


def test_pandas_scan_ignore_errors(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({"id": [1, 2, 3, 1]})
    conn.execute("CREATE NODE TABLE person(id INT64, PRIMARY KEY(id))")
    conn.execute("COPY person FROM df(IGNORE_ERRORS=true)")

    people = conn.execute("MATCH (p:person) RETURN p.id")
    assert people.get_next() == [1]
    assert people.get_next() == [2]
    assert people.get_next() == [3]
    assert not people.has_next()

    warnings = conn.execute("CALL show_warnings() RETURN *")
    assert warnings.get_next()[1].startswith("Found duplicated primary key value 1")
    assert not warnings.has_next()


def test_copy_from_pandas_multi_pairs(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute("CREATE NODE TABLE person(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:person {id: 3});")
    conn.execute("CREATE (p:person {id: 4});")
    conn.execute("CREATE NODE TABLE student(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:student {id: 2});")
    conn.execute("CREATE REL TABLE knows(from person to person, from person to student, length int64)")
    df = pd.DataFrame({
        "from": [3],
        "to": [4],
        "length": [252]
    })
    conn.execute("COPY knows from df (from = 'person', to = 'person');")
    result = conn.execute("match (:person)-[e:knows]->(:person) return e.*")
    assert result.has_next()
    assert result.get_next()[0] == 252
    assert not result.has_next()


def test_scan_pandas_with_exists(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    conn.execute("CREATE NODE TABLE person(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:person {id: 1})")
    conn.execute("CREATE (p:person {id: 2})")
    conn.execute("CREATE (p:person {id: 3})")
    conn.execute("CREATE REL TABLE knows(from person to person)")
    df = pd.DataFrame({
        "from": [1, 2, 3],
        "to": [3, 2, 1],
    })
    conn.execute(
        "COPY knows from (load from df where not exists {MATCH (p:person)-[:knows]->(p1:person) WHERE p.id = from AND p1.id = to} return from + 1 - 1, to)")
    res = conn.execute("MATCH (p:person)-[:knows]->(p1:person) return p.id, p1.id order by p.id, p1.id")
    assert res.has_next()
    tp = res.get_next()
    assert tp[0] == 1
    assert tp[1] == 3
    tp = res.get_next()
    assert tp[0] == 2
    assert tp[1] == 2
    tp = res.get_next()
    assert tp[0] == 3
    assert tp[1] == 1

def test_scan_empty_list(tmp_path: Path) -> None:
    db = kuzu.Database(tmp_path)
    conn = kuzu.Connection(db)
    df = pd.DataFrame({
        "id": ["1"],
        "lstcol": [[]]
    })
    res = conn.execute("load from df return *")
    assert res.has_next()
    tp = res.get_next()
    assert tp[0] == "1"
    assert tp[1] == []



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .torch_geometric_result_converter import TorchGeometricResultConverter
from .types import Type

if TYPE_CHECKING:
    import sys
    from types import TracebackType

    import networkx as nx
    import pandas as pd
    import polars as pl
    import pyarrow as pa
    import torch_geometric.data as geo

    from . import _kuzu

    if sys.version_info >= (3, 11):
        from typing import Self
    else:
        from typing_extensions import Self


class QueryResult:
    """QueryResult stores the result of a query execution."""

    def __init__(self, connection: _kuzu.Connection, query_result: _kuzu.QueryResult):  # type: ignore[name-defined]
        """
        Parameters
        ----------
        connection : _kuzu.Connection
            The underlying C++ connection object from pybind11.

        query_result : _kuzu.QueryResult
            The underlying C++ query result object from pybind11.

        """
        self.connection = connection
        self._query_result = query_result
        self.is_closed = False

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        self.close()

    def __del__(self) -> None:
        self.close()

    def check_for_query_result_close(self) -> None:
        """
        Check if the query result is closed and raise an exception if it is.

        Raises
        ------
        Exception
            If the query result is closed.

        """
        if self.is_closed:
            msg = "Query result is closed"
            raise RuntimeError(msg)

    def has_next(self) -> bool:
        """
        Check if there are more rows in the query result.

        Returns
        -------
        bool
            True if there are more rows in the query result, False otherwise.

        """
        self.check_for_query_result_close()
        return self._query_result.hasNext()

    def get_next(self) -> list[Any]:
        """
        Get the next row in the query result.

        Returns
        -------
        list
            Next row in the query result.

        """
        self.check_for_query_result_close()
        return self._query_result.getNext()

    def close(self) -> None:
        """Close the query result."""
        if not self.is_closed:
            # Allows the connection to be garbage collected if the query result
            # is closed manually by the user.
            self._query_result.close()
            self.connection = None
            self.is_closed = True

    def get_as_df(self) -> pd.DataFrame:
        """
        Get the query result as a Pandas DataFrame.

        See Also
        --------
        get_as_pl : Get the query result as a Polars DataFrame.
        get_as_arrow : Get the query result as a PyArrow Table.

        Returns
        -------
        pandas.DataFrame
            Query result as a Pandas DataFrame.

        """
        self.check_for_query_result_close()

        return self._query_result.getAsDF()

    def get_as_pl(self) -> pl.DataFrame:
        """
        Get the query result as a Polars DataFrame.

        See Also
        --------
        get_as_df : Get the query result as a Pandas DataFrame.
        get_as_arrow : Get the query result as a PyArrow Table.

        Returns
        -------
        polars.DataFrame
            Query result as a Polars DataFrame.
        """
        import polars as pl

        self.check_for_query_result_close()

        # note: polars should always export just a single chunk,
        # (eg: "-1") otherwise it will just need to rechunk anyway
        return pl.from_arrow(  # type: ignore[return-value]
            data=self.get_as_arrow(chunk_size=-1),
        )

    def get_as_arrow(self, chunk_size: int | None = None) -> pa.Table:
        """
        Get the query result as a PyArrow Table.

        Parameters
        ----------
        chunk_size : Number of rows to include in each chunk.
            None
                The chunk size is adaptive and depends on the number of columns in the query result.
            -1 or 0
                The entire result is returned as a single chunk.
            > 0
                The chunk size is the number of rows specified.

        See Also
        --------
        get_as_pl : Get the query result as a Polars DataFrame.
        get_as_df : Get the query result as a Pandas DataFrame.

        Returns
        -------
        pyarrow.Table
            Query result as a PyArrow Table.
        """
        self.check_for_query_result_close()

        if chunk_size is None:
            # Adaptive; target 10m total elements in each chunk.
            # (eg: if we had 10 cols, this would result in a 1m row chunk_size).
            target_n_elems = 10_000_000
            chunk_size = max(target_n_elems // len(self.get_column_names()), 10)
        elif chunk_size <= 0:
            # No chunking: return the entire result as a single chunk
            chunk_size = self.get_num_tuples()

        return self._query_result.getAsArrow(chunk_size)

    def get_column_data_types(self) -> list[str]:
        """
        Get the data types of the columns in the query result.

        Returns
        -------
        list
            Data types of the columns in the query result.

        """
        self.check_for_query_result_close()
        return self._query_result.getColumnDataTypes()

    def get_column_names(self) -> list[str]:
        """
        Get the names of the columns in the query result.

        Returns
        -------
        list
            Names of the columns in the query result.

        """
        self.check_for_query_result_close()
        return self._query_result.getColumnNames()

    def get_schema(self) -> dict[str, str]:
        """
        Get the column schema of the query result.

        Returns
        -------
        dict
            Schema of the query result.

        """
        self.check_for_query_result_close()
        return dict(
            zip(
                self._query_result.getColumnNames(),
                self._query_result.getColumnDataTypes(),
            )
        )

    def reset_iterator(self) -> None:
        """Reset the iterator of the query result."""
        self.check_for_query_result_close()
        self._query_result.resetIterator()

    def get_as_networkx(
        self,
        directed: bool = True,  # noqa: FBT001
    ) -> nx.MultiGraph | nx.MultiDiGraph:
        """
        Convert the nodes and rels in query result into a NetworkX directed or undirected graph
        with the following rules:
        Columns with data type other than node or rel will be ignored.
        Duplicated nodes and rels will be converted only once.

        Parameters
        ----------
        directed : bool
            Whether the graph should be directed. Defaults to True.

        Returns
        -------
        networkx.MultiDiGraph or networkx.MultiGraph
            Query result as a NetworkX graph.

        """
        self.check_for_query_result_close()
        import networkx as nx

        nx_graph = nx.MultiDiGraph() if directed else nx.MultiGraph()
        properties_to_extract = self._get_properties_to_extract()

        self.reset_iterator()

        nodes = {}
        rels = {}
        table_to_label_dict = {}
        table_primary_key_dict = {}

        def encode_node_id(node: dict[str, Any], table_primary_key_dict: dict[str, Any]) -> str:
            node_label = node["_label"]
            return f"{node_label}_{node[table_primary_key_dict[node_label]]!s}"

        def encode_rel_id(rel: dict[str, Any]) -> tuple[int, int]:
            return rel["_id"]["table"], rel["_id"]["offset"]

        # De-duplicate nodes and rels
        while self.has_next():
            row = self.get_next()
            for i in properties_to_extract:
                # Skip empty nodes and rels, which may be returned by
                # OPTIONAL MATCH
                if row[i] is None or row[i] == {}:
                    continue
                column_type, _ = properties_to_extract[i]
                if column_type == Type.NODE.value:
                    _id = row[i]["_id"]
                    nodes[(_id["table"], _id["offset"])] = row[i]
                    table_to_label_dict[_id["table"]] = row[i]["_label"]

                elif column_type == Type.REL.value:
                    _src = row[i]["_src"]
                    _dst = row[i]["_dst"]
                    rels[encode_rel_id(row[i])] = row[i]

                elif column_type == Type.RECURSIVE_REL.value:
                    for node in row[i]["_nodes"]:
                        _id = node["_id"]
                        nodes[(_id["table"], _id["offset"])] = node
                        table_to_label_dict[_id["table"]] = node["_label"]
                    for rel in row[i]["_rels"]:
                        for key in list(rel.keys()):
                            if rel[key] is None:
                                del rel[key]
                        _src = rel["_src"]
                        _dst = rel["_dst"]
                        rels[encode_rel_id(rel)] = rel

        # Add nodes
        for node in nodes.values():
            _id = node["_id"]
            node_id = node["_label"] + "_" + str(_id["offset"])
            if node["_label"] not in table_primary_key_dict:
                props = self.connection._get_node_property_names(node["_label"])
                for prop_name in props:
                    if props[prop_name]["is_primary_key"]:
                        table_primary_key_dict[node["_label"]] = prop_name
                        break
            node_id = encode_node_id(node, table_primary_key_dict)
            node[node["_label"]] = True
            nx_graph.add_node(node_id, **node)

        # Add rels
        for rel in rels.values():
            _src = rel["_src"]
            _dst = rel["_dst"]
            src_node = nodes[(_src["table"], _src["offset"])]
            dst_node = nodes[(_dst["table"], _dst["offset"])]
            src_id = encode_node_id(src_node, table_primary_key_dict)
            dst_id = encode_node_id(dst_node, table_primary_key_dict)
            nx_graph.add_edge(src_id, dst_id, **rel)
        return nx_graph

    def _get_properties_to_extract(self) -> dict[int, tuple[str, str]]:
        column_names = self.get_column_names()
        column_types = self.get_column_data_types()
        properties_to_extract = {}

        # Iterate over columns and extract nodes and rels, ignoring other columns
        for i in range(len(column_names)):
            column_name = column_names[i]
            column_type = column_types[i]
            if column_type in [
                Type.NODE.value,
                Type.REL.value,
                Type.RECURSIVE_REL.value,
            ]:
                properties_to_extract[i] = (column_type, column_name)
        return properties_to_extract

    def get_as_torch_geometric(self) -> tuple[geo.Data | geo.HeteroData, dict, dict, dict]:  # type: ignore[type-arg]
        """
        Converts the nodes and rels in query result into a PyTorch Geometric graph representation
        torch_geometric.data.Data or torch_geometric.data.HeteroData.

        For node conversion, numerical and boolean properties are directly converted into tensor and
        stored in Data/HeteroData. For properties cannot be converted into tensor automatically
        (please refer to the notes below for more detail), they are returned as unconverted_properties.

        For rel conversion, rel is converted into edge_index tensor director. Edge properties are returned
        as edge_properties.

        Node properties that cannot be converted into tensor automatically:
        - If the type of a node property is not one of INT64, DOUBLE, or BOOL, it cannot be converted
          automatically.
        - If a node property contains a null value, it cannot be converted automatically.
        - If a node property contains a nested list of variable length (e.g. [[1,2],[3]]), it cannot be
          converted automatically.
        - If a node property is a list or nested list, but the shape is inconsistent (e.g. the list length
          is 6 for one node but 5 for another node), it cannot be converted automatically.

        Additional conversion rules:
        - Columns with data type other than node or rel will be ignored.
        - Duplicated nodes and rels will be converted only once.

        Returns
        -------
        torch_geometric.data.Data or torch_geometric.data.HeteroData
            Query result as a PyTorch Geometric graph. Containing numeric or boolean node properties
            and edge_index tensor.

        dict
            A dictionary that maps the positional offset of each node in Data/HeteroData to its primary
            key in the database.

        dict
            A dictionary contains node properties that cannot be converted into tensor automatically. The
            order of values for each property is aligned with nodes in Data/HeteroData.

        dict
            A dictionary contains edge properties. The order of values for each property is aligned with
            edge_index in Data/HeteroData.
        """
        self.check_for_query_result_close()
        # Despite we are not using torch_geometric in this file, we need to
        # import it here to throw an error early if the user does not have
        # torch_geometric or torch installed.

        converter = TorchGeometricResultConverter(self)
        return converter.get_as_torch_geometric()

    def get_execution_time(self) -> int:
        """
        Get the time in ms which was required for executing the query.

        Returns
        -------
        double
            Query execution time as double in ms.

        """
        self.check_for_query_result_close()
        return self._query_result.getExecutionTime()

    def get_compiling_time(self) -> int:
        """
        Get the time in ms which was required for compiling the query.

        Returns
        -------
        double
            Query compile time as double in ms.

        """
        self.check_for_query_result_close()
        return self._query_result.getCompilingTime()

    def get_num_tuples(self) -> int:
        """
        Get the number of tuples which the query returned.

        Returns
        -------
        int
            Number of tuples.

        """
        self.check_for_query_result_close()
        return self._query_result.getNumTuples()



"""
# Kuzu Python API bindings.

This package provides a Python API for Kuzu graph database management system.

To install the package, run:
```
python3 -m pip install kuzu
```

Example usage:
```python
import kuzu

db = kuzu.Database("./test")
conn = kuzu.Connection(db)

# Define the schema
conn.execute("CREATE NODE TABLE User(name STRING, age INT64, PRIMARY KEY (name))")
conn.execute("CREATE NODE TABLE City(name STRING, population INT64, PRIMARY KEY (name))")
conn.execute("CREATE REL TABLE Follows(FROM User TO User, since INT64)")
conn.execute("CREATE REL TABLE LivesIn(FROM User TO City)")

# Load some data
conn.execute('COPY User FROM "user.csv"')
conn.execute('COPY City FROM "city.csv"')
conn.execute('COPY Follows FROM "follows.csv"')
conn.execute('COPY LivesIn FROM "lives-in.csv"')

# Query the data
results = conn.execute("MATCH (u:User) RETURN u.name, u.age;")
while results.has_next():
    print(results.get_next())
```

The dataset used in this example can be found [here](https://github.com/kuzudb/kuzu/tree/master/dataset/demo-db/csv).

"""

from __future__ import annotations

import os
import sys

# Set RTLD_GLOBAL and RTLD_LAZY flags on Linux to fix the issue with loading
# extensions
if sys.platform == "linux":
    original_dlopen_flags = sys.getdlopenflags()
    sys.setdlopenflags(os.RTLD_GLOBAL | os.RTLD_LAZY)

from .async_connection import AsyncConnection
from .connection import Connection
from .database import Database
from .prepared_statement import PreparedStatement
from .query_result import QueryResult
from .types import Type


def __getattr__(name: str) -> str | int:
    if name in ("version", "__version__"):
        return Database.get_version()
    elif name == "storage_version":
        return Database.get_storage_version()
    else:
        msg = f"module {__name__!r} has no attribute {name!r}"
        raise AttributeError(msg)


# Restore the original dlopen flags
if sys.platform == "linux":
    sys.setdlopenflags(original_dlopen_flags)

__all__ = [
    "AsyncConnection",
    "Connection",
    "Database",
    "PreparedStatement",
    "QueryResult",
    "Type",
    "__version__",
    "storage_version",
    "version",
]



import subprocess
import sys
from pathlib import Path
from textwrap import dedent

from test_helper import KUZU_ROOT


def test_query_result_close(tmp_path: Path, build_dir: Path) -> None:
    code = dedent(f"""
        import sys
        sys.path.append(r"{build_dir!s}")

        import kuzu
        db = kuzu.Database(r"{tmp_path!s}")
        conn = kuzu.Connection(db)
        conn.execute('''
          CREATE NODE TABLE person (
            ID INT64,
            fName STRING,
            gender INT64,
            isStudent BOOLEAN,
            isWorker BOOLEAN,
            age INT64,
            eyeSight DOUBLE,
            birthdate DATE,
            registerTime TIMESTAMP,
            lastJobDuration INTERVAL,
            workedHours INT64[],
            usedNames STRING[],
            courseScoresPerTerm INT64[][],
            grades INT64[4],
            height float,
            u UUID,
            PRIMARY KEY (ID))
        ''')
        conn.execute('COPY person FROM "{KUZU_ROOT}/dataset/tinysnb/vPerson.csv" (HEADER=true)')
        result = conn.execute("MATCH (a:person) WHERE a.ID = 0 RETURN a.isStudent;")
        # result.close()
    """)
    result = subprocess.run([sys.executable, "-c", code])
    assert result.returncode == 0



import os

import pexpect
import pytest

from conftest import ShellTest
from test_helper import deleteIfExists


def test_basic(temp_db) -> None:
    test = ShellTest().add_argument(temp_db).statement('RETURN "databases rule" AS a;')
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")


def test_range(temp_db) -> None:
    test = ShellTest().add_argument(temp_db).statement("RETURN RANGE(0, 10) AS a;")
    result = test.run()
    result.check_stdout("[0,1,2,3,4,5,6,7,8,9,10]")


@pytest.mark.parametrize(
    ("input", "output"),
    [
        ("RETURN LIST_CREATION(1,2);", "[1,2]"),
        ("RETURN STRUCT_PACK(x:=2,y:=3);", "{x: 2, y: 3}"),
        ("RETURN STRUCT_PACK(x:=2,y:=LIST_CREATION(1,2));", "{x: 2, y: [1,2]}"),
    ],
)
def test_nested_types(temp_db, input, output) -> None:
    test = ShellTest().add_argument(temp_db).statement(input)
    result = test.run()
    result.check_stdout(output)


def test_invalid_cast(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement("CREATE NODE TABLE a(i STRING, PRIMARY KEY(i));")
        .statement("CREATE (:a {i: '****'});")
        .statement('MATCH (t:a) RETURN CAST(t.i, "INT8");')
    )
    result = test.run()
    result.check_stdout(
        'Error: Conversion exception: Cast failed. Could not convert "****" to INT8.',
    )


def test_multiline(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement("RETURN")
        .statement('"databases rule"')
        .statement("AS")
        .statement("a")
        .statement(";")
    )
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")


def test_multi_queries_one_line(temp_db) -> None:
    # two successful queries
    test = ShellTest().add_argument(temp_db).statement('RETURN "databases rule" AS a; RETURN "kuzu is cool" AS b;')
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")
    result.check_stdout("\u2502 kuzu is cool \u2502")

    # one success one failure
    test = ShellTest().add_argument(temp_db).statement('RETURN "databases rule" AS a; RETURN s;')
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")
    result.check_stdout("Error: Binder exception: Variable s is not in scope.")

    # two failing queries
    test = ShellTest().add_argument(temp_db).statement('RETURN "databases rule" S a; RETURN s;')
    result = test.run()
    result.check_stdout(
        [
            "Error: Parser exception: Invalid input < S>: expected rule ku_Statements (line: 1, offset: 24)",
            '"RETURN "databases rule" S a; RETURN s;"',
            "                         ^",
        ],
    )


def test_row_truncation(temp_db, csv_path) -> None:
    test = ShellTest().add_argument(temp_db).statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN id;')
    result = test.run()
    result.check_stdout("(21 tuples, 20 shown)")
    result.check_stdout(["\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502", "\u2502  \u00B7    \u2502"])


def test_column_truncation(temp_db, csv_path) -> None:
    # width when running test is 80
    test = ShellTest().add_argument(temp_db).statement(f'LOAD FROM "{csv_path}" (HEADER=true) RETURN *;')
    result = test.run()
    result.check_stdout("id")
    result.check_stdout("fname")
    result.check_stdout("Gender")
    result.check_stdout("\u2502 ... \u2502")
    result.check_stdout("courseScoresPerTerm")
    result.check_stdout("(13 columns, 4 shown)")

    # test column name truncation
    long_name = "a" * 100
    test = ShellTest().add_argument(temp_db).statement(f'RETURN "databases rule" AS {long_name};')
    result = test.run()
    result.check_stdout(f"\u2502 {long_name[:73]}... \u2502")
    result.check_stdout("\u2502 databases rule")


def long_messages(temp_db) -> None:
    long_message = "-" * 4096
    test = ShellTest().add_argument(temp_db).statement(f'RETURN "{long_message}" AS a;')
    result = test.run()
    result.check_stdout(long_message)


def test_history_consecutive_repeats(temp_db, history_path) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .add_argument("-p")
        .add_argument(history_path)
        .statement('RETURN "databases rule" AS a;')
        .statement('RETURN "databases rule" AS a;')
    )
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")

    with open(os.path.join(history_path, "history.txt")) as f:
        assert f.readline() == 'RETURN "databases rule" AS a;\n'
        assert f.readline() == ""

    deleteIfExists(os.path.join(history_path, "history.txt"))


def test_kuzurc(temp_db) -> None:
    deleteIfExists(".kuzurc")
    # confirm that nothing is read on startup
    test = ShellTest().add_argument(temp_db)
    result = test.run()
    result.check_not_stdout("-- Loading resources from .kuzurc")

    # create a .kuzurc file
    with open(".kuzurc", "w") as f:
        f.write("CREATE NODE TABLE a(i STRING, PRIMARY KEY(i));\n")
        f.write(":max_rows 1\n")
    
    # confirm that the file is read on startup
    test = ShellTest().add_argument(temp_db).statement("CALL show_tables() RETURN *;")
    result = test.run()
    deleteIfExists(".kuzurc")
    result.check_stdout("-- Loading resources from .kuzurc")
    result.check_stdout("maxRows set as 1")
    result.check_stdout("a")

    # create a .kuzurc file with errors
    with open(".kuzurc", "w") as f:
        f.write('RETURN "databases rule" S a; RETURN s;\n')
        f.write(":max_rows\n")
        f.write(":mode table\n")
        f.write("CREATE NODE TABLE b(i STRING, PRIMARY KEY(i));\n")

    # confirm that the file is read on startup
    test = ShellTest().add_argument(temp_db).statement("CALL show_tables() RETURN *;")
    result = test.run()
    deleteIfExists(".kuzurc")
    result.check_stdout("-- Loading resources from .kuzurc")
    result.check_stdout(
        [
            "Error: Parser exception: Invalid input < S>: expected rule ku_Statements (line: 1, offset: 24)",
            '"RETURN "databases rule" S a; RETURN s;"',
            "                         ^",
        ],
    )
    result.check_stdout("Cannot parse '' as number of rows. Expect integer.")
    result.check_stdout("mode set as table")
    result.check_stdout("b")


def test_comments(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement('RETURN // testing\n /* test\ntest\ntest */"databases rule" // testing\n AS a\n; // testing')
        .statement('\x1b[A\r') # run the last command again to make sure comments are still ignored
    )
    result = test.run()
    result.check_stdout("\u2502 databases rule \u2502")


def test_shell_auto_completion(temp_db) -> None:
    test = ShellTest().add_argument(temp_db)
    test.start()
    test.send_statement("CREATE NODE TABLE coolTable(i STRING, longPropertyName STRING, autoCompleteUINT8 UIN\t")
    test.send_statement("\t")
    test.send_statement("\t")
    test.send_statement("\t")
    test.send_statement(", PRIMARY KEY(i));\n")
    assert test.shell_process.expect_exact(["\u2502 Table coolTable has been created. \u2502", pexpect.EOF]) == 0
    
    test.send_statement("ma\t")
    test.send_statement("\t")
    test.send_statement("\t")
    test.send_statement("\x1b[Z")
    test.send_statement(" (n:\t")
    test.send_statement(") ret\t")
    test.send_statement(" n.\t")
    test.send_statement("\t")
    test.send_finished_statement(";\r")
    assert test.shell_process.expect_exact(["\u2502 n.longPropertyName \u2502", pexpect.EOF]) == 0

    test.send_statement("c\t")
    test.send_statement(" show_t\t")
    test.send_statement("() ret\t")
    test.send_finished_statement(" *;\n")
    assert test.shell_process.expect_exact(["\u2502 coolTable \u2502", pexpect.EOF]) == 0


def test_shell_unicode_input(temp_db) -> None:
    test = (
        ShellTest()
        .add_argument(temp_db)
        .statement('CREATE NODE TABLE IF NOT EXISTS `B\\u00fccher` (title STRING, price INT64, PRIMARY KEY (title));\n') 
        .statement("CREATE (n:`B\\u00fccher` {title: 'Der Thron der Sieben K\\00f6nigreiche'}) SET n.price = 20;\n")
        .statement('MATCH (n:B\\u00fccher) RETURN label(n);\n')
        .statement('return "\\uD83D\\uDE01";\n') # surrogate pair for grinning face emoji
        .statement('return "\\U0001F601";\n') # grinning face emoji
        .statement('return "\\uD83D";\n') #  unmatched surrogate pair
        .statement('return "\\uDE01";\n') # unmatched surrogate pair
        .statement('return "\\uD83D\\uDBFF";\n') # bad lower surrogate
        .statement('return "\\u000";\n') # bad unicode codepoint
        .statement('return "\\u0000";\n') # Null character
        .statement('return "\\U00110000";\n') # Invalid codepoint
    )
    result = test.run()
    result.check_stdout("\u2502 B\u00fccher")
    result.check_stdout("\u2502 \U0001F601") # grinning face emoji
    result.check_stdout("\u2502 \U0001F601") # grinning face emoji
    result.check_stdout("Error: Unmatched high surrogate")
    result.check_stdout("Error: Failed to convert codepoint to UTF-8")
    result.check_stdout("Error: Invalid surrogate pair")
    result.check_stdout("Error: Parser exception: Invalid input <return \">: expected rule oC_RegularQuery (line: 1, offset: 7)")
    result.check_stdout("Error: Null character not allowed")
    result.check_stdout("Error: Invalid Unicode codepoint")



from __future__ import annotations

import datetime
import uuid

from type_aliases import ConnDB


def test_read(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    prepared_statement = conn.prepare("MATCH (a:person) WHERE a.isStudent = $1 AND a.isWorker = $k RETURN COUNT(*)")
    assert prepared_statement.is_success()
    assert prepared_statement.get_error_message() == ""

    result = conn.execute(prepared_statement, {"1": False, "k": False})
    assert result.has_next()
    assert result.get_next() == [1]
    assert not result.has_next()

    result = conn.execute(prepared_statement, {"1": True, "k": False})
    assert result.has_next()
    assert result.get_next() == [3]
    assert not result.has_next()

    result = conn.execute(prepared_statement, {"1": False, "k": True})
    assert result.has_next()
    assert result.get_next() == [4]
    assert not result.has_next()

    result = conn.execute(prepared_statement, {"1": True, "k": True})
    assert result.has_next()
    assert result.get_next() == [0]
    assert not result.has_next()


def test_null_value(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    prepared_statement = conn.prepare("RETURN [4, $1, 2, $3, 4]")
    assert prepared_statement.is_success()
    assert prepared_statement.get_error_message() == ""

    result = conn.execute(prepared_statement, {"1": None, "3": 5})
    assert result.has_next()
    assert result.get_next() == [[4, None, 2, 5, 4]]
    assert not result.has_next()


def test_write(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite
    orgs = [
        {
            "ID": 1001,
            "name": "org1",
            "orgCode": 1,
            "mark": 1.0,
            "score": 1,
            "history": "history1",
            "licenseValidInterval": datetime.timedelta(days=1),
            "rating": 1.0,
        },
        {
            "ID": 1002,
            "name": "org2",
            "orgCode": 2,
            "mark": 2.0,
            "score": 2,
            "history": "history2",
            "licenseValidInterval": datetime.timedelta(days=2),
            "rating": 2.0,
        },
        {
            "ID": 1003,
            "name": "org3",
            "orgCode": 3,
            "mark": 3.0,
            "score": 3,
            "history": "history3",
            "licenseValidInterval": datetime.timedelta(days=3),
            "rating": 3.0,
        },
    ]

    prepared_statement = conn.prepare(
        "CREATE (n:organisation {ID: $ID, name: $name, orgCode: $orgCode, mark: $mark, score: $score, history: $history, licenseValidInterval: $licenseValidInterval, rating: $rating})"
    )
    assert prepared_statement.is_success()
    for org in orgs:
        org_dict = {str(k): v for k, v in org.items()}
        conn.execute(prepared_statement, org_dict)

    all_orgs_res = conn.execute("MATCH (n:organisation) RETURN n")
    while all_orgs_res.has_next():
        n = all_orgs_res.get_next()[0]
        if n["ID"] not in [o["ID"] for o in orgs]:
            continue
        for expected_org in orgs:
            if n["ID"] == expected_org["ID"]:
                assert n["ID"] == expected_org["ID"]
                assert n["name"] == expected_org["name"]
                assert n["orgCode"] == expected_org["orgCode"]
                assert n["mark"] == expected_org["mark"]
                assert n["score"] == expected_org["score"]
                assert n["history"] == expected_org["history"]
                assert n["licenseValidInterval"] == expected_org["licenseValidInterval"]
                assert n["rating"] == expected_org["rating"]
                break

    conn.execute("CREATE NODE TABLE uuid_table (id UUID, PRIMARY KEY(id));")
    conn.execute("CREATE (:uuid_table {id: $1});", {"1": uuid.uuid5(uuid.NAMESPACE_DNS, "kuzu")})
    result = conn.execute("MATCH (n:uuid_table) RETURN n.id;")
    assert result.get_next() == [uuid.uuid5(uuid.NAMESPACE_DNS, "kuzu")]


def test_error(conn_db_readonly: ConnDB) -> None:
    prepared_statement = conn_db_readonly[0].prepare("MATCH (d:dog) WHERE d.isServiceDog = $1 RETURN COUNT(*)")
    assert not prepared_statement.is_success()
    assert prepared_statement.get_error_message() == "Binder exception: Table dog does not exist."

    prepared_statement = conn_db_readonly[0].prepare("SELECT * FROM person")
    assert not prepared_statement.is_success()
    assert prepared_statement.get_error_message().startswith("Parser exception: extraneous input 'SELECT'")

def test_prepare_limit(conn_db_readonly: ConnDB) -> None:
    prepared_statement = conn_db_readonly[0].prepare("MATCH (p:person) return p.ID limit $lt")
    result = conn_db_readonly[0].execute(prepared_statement, {'lt': 3})
    assert result.get_next() == [0]
    assert result.get_next() == [2]
    assert result.get_next() == [3]



from __future__ import annotations

import warnings

import ground_truth
import torch
from type_aliases import ConnDB


def test_to_torch_geometric_nodes_only(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:person) return p"

    res = conn.execute(query)
    with warnings.catch_warnings(record=True) as ws:
        torch_geometric_data, pos_to_idx, unconverted_properties, _ = res.get_as_torch_geometric()
    warnings_ground_truth = {
        "Property person.courseScoresPerTerm cannot be converted to Tensor (likely due to nested list of variable length). The property is marked as unconverted.",
        "Property person.height of type FLOAT is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.u of type UUID is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.lastJobDuration of type INTERVAL is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.registerTime of type TIMESTAMP is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.birthdate of type DATE is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.fName of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.workedHours has an inconsistent shape. The property is marked as unconverted.",
        "Property person.usedNames of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
    }
    assert len(ws) == 9
    for w in ws:
        assert str(w.message) in warnings_ground_truth

    assert torch_geometric_data.ID.shape == torch.Size([8])
    assert torch_geometric_data.ID.dtype == torch.int64
    for i in range(8):
        assert ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["ID"] == torch_geometric_data.ID[i].item()

    assert torch_geometric_data.gender.shape == torch.Size([8])
    assert torch_geometric_data.gender.dtype == torch.int64
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["gender"]
            == torch_geometric_data.gender[i].item()
        )

    assert torch_geometric_data.isStudent.shape == torch.Size([8])
    assert torch_geometric_data.isStudent.dtype == torch.bool
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["isStudent"]
            == torch_geometric_data.isStudent[i].item()
        )

    assert torch_geometric_data.isWorker.shape == torch.Size([8])
    assert torch_geometric_data.isWorker.dtype == torch.bool
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["isWorker"]
            == torch_geometric_data.isWorker[i].item()
        )

    assert torch_geometric_data.age.shape == torch.Size([8])
    assert torch_geometric_data.age.dtype == torch.int64
    for i in range(8):
        assert ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["age"] == torch_geometric_data.age[i].item()

    assert torch_geometric_data.eyeSight.shape == torch.Size([8])
    assert torch_geometric_data.eyeSight.dtype == torch.float32
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["eyeSight"]
            - torch_geometric_data.eyeSight[i].item()
            < 1e-6
        )

    assert len(unconverted_properties) == 9
    assert "courseScoresPerTerm" in unconverted_properties
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["courseScoresPerTerm"]
            == unconverted_properties["courseScoresPerTerm"][i]
        )
    assert "lastJobDuration" in unconverted_properties
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["lastJobDuration"]
            == unconverted_properties["lastJobDuration"][i]
        )
    assert "registerTime" in unconverted_properties
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["registerTime"]
            == unconverted_properties["registerTime"][i]
        )
    assert "birthdate" in unconverted_properties
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["birthdate"]
            == unconverted_properties["birthdate"][i]
        )
    assert "fName" in unconverted_properties
    for i in range(8):
        assert ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["fName"] == unconverted_properties["fName"][i]
    assert "usedNames" in unconverted_properties
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["usedNames"]
            == unconverted_properties["usedNames"][i]
        )

    assert "workedHours" in unconverted_properties
    for i in range(8):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["workedHours"]
            == unconverted_properties["workedHours"][i]
        )


def test_to_torch_geometric_homogeneous_graph(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:person)-[r:knows]->(q:person) RETURN p, r, q"

    res = conn.execute(query)
    with warnings.catch_warnings(record=True) as ws:
        torch_geometric_data, pos_to_idx, unconverted_properties, edge_properties = res.get_as_torch_geometric()
    warnings_ground_truth = {
        "Property person.courseScoresPerTerm cannot be converted to Tensor (likely due to nested list of variable length). The property is marked as unconverted.",
        "Property person.height of type FLOAT is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.u of type UUID is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.lastJobDuration of type INTERVAL is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.registerTime of type TIMESTAMP is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.birthdate of type DATE is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.fName of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.workedHours has an inconsistent shape. The property is marked as unconverted.",
        "Property person.usedNames of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
    }
    assert len(ws) == 9
    for w in ws:
        assert str(w.message) in warnings_ground_truth

    assert torch_geometric_data.ID.shape == torch.Size([7])
    assert torch_geometric_data.ID.dtype == torch.int64
    for i in range(7):
        assert ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["ID"] == torch_geometric_data.ID[i].item()

    assert torch_geometric_data.gender.shape == torch.Size([7])
    assert torch_geometric_data.gender.dtype == torch.int64
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["gender"]
            == torch_geometric_data.gender[i].item()
        )

    assert torch_geometric_data.isStudent.shape == torch.Size([7])
    assert torch_geometric_data.isStudent.dtype == torch.bool
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["isStudent"]
            == torch_geometric_data.isStudent[i].item()
        )

    assert torch_geometric_data.isWorker.shape == torch.Size([7])
    assert torch_geometric_data.isWorker.dtype == torch.bool
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["isWorker"]
            == torch_geometric_data.isWorker[i].item()
        )

    assert torch_geometric_data.age.shape == torch.Size([7])
    assert torch_geometric_data.age.dtype == torch.int64
    for i in range(7):
        assert ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["age"] == torch_geometric_data.age[i].item()

    assert torch_geometric_data.eyeSight.shape == torch.Size([7])
    assert torch_geometric_data.eyeSight.dtype == torch.float32
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["eyeSight"]
            - torch_geometric_data.eyeSight[i].item()
            < 1e-6
        )

    assert len(unconverted_properties) == 9
    assert "courseScoresPerTerm" in unconverted_properties
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["courseScoresPerTerm"]
            == unconverted_properties["courseScoresPerTerm"][i]
        )
    assert "lastJobDuration" in unconverted_properties
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["lastJobDuration"]
            == unconverted_properties["lastJobDuration"][i]
        )
    assert "registerTime" in unconverted_properties
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["registerTime"]
            == unconverted_properties["registerTime"][i]
        )
    assert "birthdate" in unconverted_properties
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["birthdate"]
            == unconverted_properties["birthdate"][i]
        )
    assert "fName" in unconverted_properties
    for i in range(7):
        assert ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["fName"] == unconverted_properties["fName"][i]
    assert "usedNames" in unconverted_properties
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["usedNames"]
            == unconverted_properties["usedNames"][i]
        )

    assert "workedHours" in unconverted_properties
    for i in range(7):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx[i]]["workedHours"]
            == unconverted_properties["workedHours"][i]
        )

    assert torch_geometric_data.edge_index.shape == torch.Size([2, 14])
    for i in range(14):
        src, dst = torch_geometric_data.edge_index[0][i].item(), torch_geometric_data.edge_index[1][i].item()
        assert src in pos_to_idx
        assert dst in pos_to_idx
        assert src != dst
        assert pos_to_idx[dst] in ground_truth.TINY_SNB_KNOWS_GROUND_TRUTH[pos_to_idx[src]]

    assert len(edge_properties) == 8
    assert "date" in edge_properties
    assert "meetTime" in edge_properties
    assert "validInterval" in edge_properties
    assert "comments" in edge_properties
    assert "summary" in edge_properties
    assert "notes" in edge_properties

    for i in range(14):
        src, dst = torch_geometric_data.edge_index[0][i].item(), torch_geometric_data.edge_index[1][i].item()
        orginal_src = pos_to_idx[src]
        orginal_dst = pos_to_idx[dst]
        assert (orginal_src, orginal_dst) in ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["date"]
            == edge_properties["date"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["meetTime"]
            == edge_properties["meetTime"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["validInterval"]
            == edge_properties["validInterval"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["comments"]
            == edge_properties["comments"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["summary"]
            == edge_properties["summary"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["notes"]
            == edge_properties["notes"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(orginal_src, orginal_dst)]["someMap"]
            == edge_properties["someMap"][i]
        )


def test_to_torch_geometric_heterogeneous_graph(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:person)-[r1:knows]->(q:person)-[r2:workAt]->(o:organisation) RETURN p, q, o, r1, r2"

    res = conn.execute(query)
    with warnings.catch_warnings(record=True) as ws:
        torch_geometric_data, pos_to_idx, unconverted_properties, edge_properties = res.get_as_torch_geometric()

    assert len(ws) == 13
    warnings_ground_truth = {
        "Property organisation.name of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.height of type FLOAT is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.u of type UUID is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.courseScoresPerTerm cannot be converted to Tensor (likely due to nested list of variable length). The property is marked as unconverted.",
        "Property person.lastJobDuration of type INTERVAL is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.registerTime of type TIMESTAMP is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.birthdate of type DATE is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.fName of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property organisation.history of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property person.usedNames of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property organisation.licenseValidInterval of type INTERVAL is not supported by torch_geometric. The property is marked as unconverted.",
        "Property organisation.state of type STRUCT(revenue INT16, location STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property organisation.info of type UNION(price FLOAT, movein DATE, note STRING) is not supported by torch_geometric. The property is marked as unconverted.",
    }

    for w in ws:
        assert str(w.message) in warnings_ground_truth

    assert torch_geometric_data["person"].ID.shape == torch.Size([4])
    assert torch_geometric_data["person"].ID.dtype == torch.int64
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["ID"]
            == torch_geometric_data["person"].ID[i].item()
        )

    assert torch_geometric_data["person"].gender.shape == torch.Size([4])
    assert torch_geometric_data["person"].gender.dtype == torch.int64
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["gender"]
            == torch_geometric_data["person"].gender[i].item()
        )

    assert torch_geometric_data["person"].isStudent.shape == torch.Size([4])
    assert torch_geometric_data["person"].isStudent.dtype == torch.bool
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["isStudent"]
            == torch_geometric_data["person"].isStudent[i].item()
        )

    assert torch_geometric_data["person"].isWorker.shape == torch.Size([4])
    assert torch_geometric_data["person"].isWorker.dtype == torch.bool
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["isWorker"]
            == torch_geometric_data["person"].isWorker[i].item()
        )

    assert torch_geometric_data["person"].age.shape == torch.Size([4])
    assert torch_geometric_data["person"].age.dtype == torch.int64
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["age"]
            == torch_geometric_data["person"].age[i].item()
        )

    assert torch_geometric_data["person"].eyeSight.shape == torch.Size([4])
    assert torch_geometric_data["person"].eyeSight.dtype == torch.float32
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["eyeSight"]
            - torch_geometric_data["person"].eyeSight[i].item()
            < 1e-6
        )

    assert "person" in unconverted_properties
    assert len(unconverted_properties["person"]) == 8
    assert "courseScoresPerTerm" in unconverted_properties["person"]
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["courseScoresPerTerm"]
            == unconverted_properties["person"]["courseScoresPerTerm"][i]
        )
    assert "lastJobDuration" in unconverted_properties["person"]
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["lastJobDuration"]
            == unconverted_properties["person"]["lastJobDuration"][i]
        )
    assert "registerTime" in unconverted_properties["person"]
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["registerTime"]
            == unconverted_properties["person"]["registerTime"][i]
        )
    assert "birthdate" in unconverted_properties["person"]
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["birthdate"]
            == unconverted_properties["person"]["birthdate"][i]
        )
    assert "fName" in unconverted_properties["person"]
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["fName"]
            == unconverted_properties["person"]["fName"][i]
        )
    assert "usedNames" in unconverted_properties["person"]
    for i in range(4):
        assert (
            ground_truth.TINY_SNB_PERSONS_GROUND_TRUTH[pos_to_idx["person"][i]]["usedNames"]
            == unconverted_properties["person"]["usedNames"][i]
        )

    assert torch_geometric_data["person", "person"].edge_index.shape == torch.Size([2, 6])
    for i in range(3):
        src, dst = (
            torch_geometric_data["person", "person"].edge_index[0][i].item(),
            torch_geometric_data["person", "person"].edge_index[1][i].item(),
        )
        assert src in pos_to_idx["person"]
        assert dst in pos_to_idx["person"]
        assert src != dst
        assert pos_to_idx["person"][dst] in ground_truth.TINY_SNB_KNOWS_GROUND_TRUTH[pos_to_idx["person"][src]]

    assert len(edge_properties["person", "person"]) == 8
    assert "date" in edge_properties["person", "person"]
    assert "meetTime" in edge_properties["person", "person"]
    assert "validInterval" in edge_properties["person", "person"]
    assert "comments" in edge_properties["person", "person"]
    assert "summary" in edge_properties["person", "person"]
    assert "notes" in edge_properties["person", "person"]
    assert "_label" in edge_properties["person", "person"]
    for i in range(3):
        src, dst = (
            torch_geometric_data["person", "person"].edge_index[0][i].item(),
            torch_geometric_data["person", "person"].edge_index[1][i].item(),
        )
        original_src, original_dst = pos_to_idx["person"][src], pos_to_idx["person"][dst]
        assert (original_src, original_dst) in ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["date"]
            == edge_properties["person", "person"]["date"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["meetTime"]
            == edge_properties["person", "person"]["meetTime"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["validInterval"]
            == edge_properties["person", "person"]["validInterval"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["comments"]
            == edge_properties["person", "person"]["comments"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["summary"]
            == edge_properties["person", "person"]["summary"][i]
        )
        assert (
            ground_truth.TINY_SNB_KNOWS_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["notes"]
            == edge_properties["person", "person"]["notes"][i]
        )
        assert edge_properties["person", "person"]["_label"][i] == "knows"

    assert torch_geometric_data["organisation"].ID.shape == torch.Size([2])
    assert torch_geometric_data["organisation"].ID.dtype == torch.int64
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["ID"]
            == torch_geometric_data["organisation"].ID[i].item()
        )

    assert torch_geometric_data["organisation"].orgCode.shape == torch.Size([2])
    assert torch_geometric_data["organisation"].orgCode.dtype == torch.int64
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["orgCode"]
            == torch_geometric_data["organisation"].orgCode[i].item()
        )

    assert torch_geometric_data["organisation"].mark.shape == torch.Size([2])
    assert torch_geometric_data["organisation"].mark.dtype == torch.float32
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["mark"]
            - torch_geometric_data["organisation"].mark[i].item()
            < 1e-6
        )

    assert torch_geometric_data["organisation"].score.shape == torch.Size([2])
    assert torch_geometric_data["organisation"].score.dtype == torch.int64
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["score"]
            - torch_geometric_data["organisation"].score[i].item()
            < 1e-6
        )

    assert torch_geometric_data["organisation"].rating.shape == torch.Size([2])
    assert torch_geometric_data["organisation"].rating.dtype == torch.float32
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["rating"]
            - torch_geometric_data["organisation"].rating[i].item()
            < 1e-6
        )

    assert "organisation" in unconverted_properties
    assert len(unconverted_properties["organisation"]) == 5
    assert "name" in unconverted_properties["organisation"]
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["name"]
            == unconverted_properties["organisation"]["name"][i]
        )

    assert "history" in unconverted_properties["organisation"]
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["history"]
            == unconverted_properties["organisation"]["history"][i]
        )

    assert "licenseValidInterval" in unconverted_properties["organisation"]
    for i in range(2):
        assert (
            ground_truth.TINY_SNB_ORGANISATIONS_GROUND_TRUTH[pos_to_idx["organisation"][i]]["licenseValidInterval"]
            == unconverted_properties["organisation"]["licenseValidInterval"][i]
        )

    assert torch_geometric_data["person", "organisation"].edge_index.shape == torch.Size([2, 2])
    for i in range(2):
        src, dst = (
            torch_geometric_data["person", "organisation"].edge_index[0][i].item(),
            torch_geometric_data["person", "organisation"].edge_index[1][i].item(),
        )
        assert src in pos_to_idx["person"]
        assert dst in pos_to_idx["organisation"]
        assert src != dst
        assert (
            pos_to_idx["organisation"][dst] in ground_truth.TINY_SNB_WORKS_AT_GROUND_TRUTH[pos_to_idx["person"][src]]
        )
    assert len(edge_properties["person", "organisation"]) == 4
    assert "year" in edge_properties["person", "organisation"]
    assert "_label" in edge_properties["person", "organisation"]
    for i in range(2):
        src, dst = (
            torch_geometric_data["person", "organisation"].edge_index[0][i].item(),
            torch_geometric_data["person", "organisation"].edge_index[1][i].item(),
        )
        original_src, original_dst = pos_to_idx["person"][src], pos_to_idx["organisation"][dst]
        assert (
            ground_truth.TINY_SNB_WORKS_AT_PROPERTIES_GROUND_TRUTH[(original_src, original_dst)]["year"]
            == edge_properties["person", "organisation"]["year"][i]
        )
        assert edge_properties["person", "organisation"]["_label"][i] == "workAt"


def test_to_torch_geometric_multi_dimensional_lists(
    conn_db_readonly: ConnDB,
) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (t:tensor) RETURN t"

    res = conn.execute(query)
    with warnings.catch_warnings(record=True) as ws:
        torch_geometric_data, pos_to_idx, unconverted_properties, _ = res.get_as_torch_geometric()
    assert len(ws) == 1
    assert (
        str(ws[0].message)
        == "Property tensor.oneDimInt has a null value. torch_geometric does not support null values. The property is marked as unconverted."
    )

    bool_list = []
    float_list = []
    int_list = []

    for i in range(len(pos_to_idx)):
        idx = pos_to_idx[i]
        bool_list.append(ground_truth.TENSOR_LIST_GROUND_TRUTH[idx]["boolTensor"])
        float_list.append(ground_truth.TENSOR_LIST_GROUND_TRUTH[idx]["doubleTensor"])
        int_list.append(ground_truth.TENSOR_LIST_GROUND_TRUTH[idx]["intTensor"])

    bool_tensor = torch.tensor(bool_list, dtype=torch.bool)
    float_tensor = torch.tensor(float_list, dtype=torch.float32)
    int_tensor = torch.tensor(int_list, dtype=torch.int64)

    assert torch_geometric_data.ID.shape == torch.Size([len(pos_to_idx)])
    assert torch_geometric_data.ID.dtype == torch.int64
    for i in range(len(pos_to_idx)):
        assert torch_geometric_data.ID[i].item() == pos_to_idx[i]

    assert torch_geometric_data.boolTensor.shape == bool_tensor.shape
    assert torch_geometric_data.boolTensor.dtype == bool_tensor.dtype
    assert torch.all(torch_geometric_data.boolTensor == bool_tensor)

    assert torch_geometric_data.doubleTensor.shape == float_tensor.shape
    assert torch_geometric_data.doubleTensor.dtype == float_tensor.dtype
    assert torch.all(torch_geometric_data.doubleTensor == float_tensor)

    assert torch_geometric_data.intTensor.shape == int_tensor.shape
    assert torch_geometric_data.intTensor.dtype == int_tensor.dtype
    assert torch.all(torch_geometric_data.intTensor == int_tensor)

    assert len(unconverted_properties) == 1
    assert "oneDimInt" in unconverted_properties
    assert len(unconverted_properties["oneDimInt"]) == 6
    assert unconverted_properties["oneDimInt"] == [1, 2, None, None, 5, 6]


def test_to_torch_geometric_no_properties_converted(
    conn_db_readonly: ConnDB,
) -> None:
    conn, _ = conn_db_readonly
    query = "MATCH (p:personLongString)-[r:knowsLongString]->(q:personLongString) RETURN p, r, q"

    res = conn.execute(query)
    with warnings.catch_warnings(record=True) as ws:
        torch_geometric_data, pos_to_idx, unconverted_properties, _ = res.get_as_torch_geometric()
    assert len(ws) == 3
    warnings_ground_truth = {
        "Property personLongString.name of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "Property personLongString.spouse of type STRING is not supported by torch_geometric. The property is marked as unconverted.",
        "No nodes found or all node properties are not converted.",
    }
    for w in ws:
        assert str(w.message) in warnings_ground_truth
    assert torch_geometric_data["personLongString"] == {}
    assert torch_geometric_data["personLongString", "personLongString"].edge_index.shape == torch.Size([2, 1])

    for i in range(1):
        src, dst = (
            torch_geometric_data["personLongString", "personLongString"].edge_index[0][i].item(),
            torch_geometric_data["personLongString", "personLongString"].edge_index[1][i].item(),
        )
        assert src in pos_to_idx["personLongString"]
        assert dst in pos_to_idx["personLongString"]
        assert src != dst
        assert (
            pos_to_idx["personLongString"][dst]
            in ground_truth.PERSONLONGSTRING_KNOWS_GROUND_TRUTH[pos_to_idx["personLongString"][src]]
        )

    assert len(unconverted_properties) == 1
    assert len(unconverted_properties["personLongString"]) == 2

    assert "spouse" in unconverted_properties["personLongString"]
    assert len(unconverted_properties["personLongString"]["spouse"]) == 2
    for i in range(2):
        assert (
            ground_truth.PERSONLONGSTRING_GROUND_TRUTH[pos_to_idx["personLongString"][i]]["spouse"]
            == unconverted_properties["personLongString"]["spouse"][i]
        )

    assert "name" in unconverted_properties["personLongString"]
    for i in range(2):
        assert (
            ground_truth.PERSONLONGSTRING_GROUND_TRUTH[pos_to_idx["personLongString"][i]]["name"]
            == unconverted_properties["personLongString"]["name"][i]
        )



import os
import shutil
import platform

FILE_DIR = os.path.dirname(os.path.abspath(__file__))
DST_DIR = os.path.abspath(os.path.join(FILE_DIR, "..", "extension-artifacts"))
SRC_DIR = os.path.abspath(os.path.join(FILE_DIR, "..", "extension"))


def collect_exts():
    for ext in os.listdir(SRC_DIR):
        ext_build_path = os.path.abspath(os.path.join(SRC_DIR, ext, "build"))
        if not os.path.exists(ext_build_path):
            continue
        print("Found extension: " + ext)
        ext_dst_path = os.path.abspath(os.path.join(DST_DIR, ext))
        os.makedirs(ext_dst_path, exist_ok=True)
        for f in os.listdir(ext_build_path):
            if not f.endswith(".kuzu_extension"):
                continue
            ext_file_path = os.path.abspath(os.path.join(ext_build_path, f))
            shutil.copy(ext_file_path, ext_dst_path)
            print(" \tCopied: " + f, "=>", ext_dst_path)


def find_duckdb():
    if platform.system() == "Darwin":
        candidates = [
            "/usr/local/lib/libduckdb.dylib",
            "/opt/homebrew/lib/libduckdb.dylib",
        ]
    elif platform.system() == "Linux":
        candidates = [
            "/usr/local/lib/libduckdb.so",
            "/usr/local/lib64/libduckdb.so",
            "/usr/lib/libduckdb.so",
            "/usr/lib64/libduckdb.so",
        ]
    elif platform.system() == "Windows":
        candidates = [
            "C:\\Program Files\\duckdb\\build\\release\\src\\Release\\duckdb.lib"
        ]
    for candidate in candidates:
        if os.path.exists(candidate):
            return os.path.abspath(candidate)
    return None


def copy_duckdb():
    duckdb_dst_path = os.path.abspath(os.path.join(DST_DIR, "common"))
    os.makedirs(duckdb_dst_path, exist_ok=True)
    duckdb_path = find_duckdb()
    if duckdb_path is None:
        print("DuckDB not found, copying is skipped")
        return
    shutil.copy(duckdb_path, duckdb_dst_path)
    print("Copied DuckDB: " + duckdb_path, "=>", duckdb_dst_path)


def main():
    shutil.rmtree(DST_DIR, ignore_errors=True)
    os.makedirs(DST_DIR, exist_ok=True)
    collect_exts()
    copy_duckdb()


if __name__ == "__main__":
    main()



import gdb
import subprocess


def get_machine_architecture():
    output = subprocess.check_output(["cat", "/proc/cpuinfo"])
    flags_str = ""
    for line in output.decode("utf-8").split("\n"):
        line = line.strip()
        if line.startswith("flags"):
            flags_str = line
            break
    assert len(line) > 0

    # Uses the same way as _simsimd_capabilities_x86() in simsimd.h to detect supported features
    # The only simsimd functions we use in the vector index are cos/l2/l2_sq/dot which only branch for the targets 'skylake', 'haswell', and 'serial' so we only test for those
    if "avx512f" in flags_str:
        return "skylake"
    elif "avx2" in flags_str and "f16c" in flags_str and "fma" in flags_str:
        return "haswell"
    else:
        return "serial"


bp = gdb.Breakpoint(f"simsimd_l2_f32_{get_machine_architecture()}")

gdb.execute("run < scripts/simd-dispatch-test.cypher")

try:
    gdb.execute("continue")
    # we only care if the breakpoint is hit at all
    # disable it now to prevent the test from needing to execute 'continue' many times to reach completion
    bp.enabled = False
except gdb.error:
    # the program has terminated
    pass

# Check if the breakpoint was hit
if bp.hit_count == 0:
    print(
        f"Error: did not hit the expected simsimd function for machine architecture '{get_machine_architecture()}'"
    )
    gdb.execute("quit 1")

gdb.execute("quit")



from __future__ import annotations

import pytest
from type_aliases import ConnDB


def test_timeout(conn_db_readonly: ConnDB) -> None:
    conn, _ = conn_db_readonly
    conn.set_query_timeout(1000)
    with pytest.raises(RuntimeError, match="Interrupted."):
        conn.execute("UNWIND RANGE(1,100000) AS x UNWIND RANGE(1, 100000) AS y RETURN COUNT(x + y);")



from __future__ import annotations

import threading
import time
from typing import TYPE_CHECKING

import kuzu
import pytest
from type_aliases import ConnDB

if TYPE_CHECKING:
    from pathlib import Path


def test_connection_close(tmp_path: Path, build_dir: Path) -> None:
    db_path = tmp_path / "test_connection_close.kuzu"
    db = kuzu.Database(database_path=db_path, read_only=False)
    conn = kuzu.Connection(db)
    conn.close()
    assert conn.is_closed
    pytest.raises(RuntimeError, conn.execute, "RETURN 1")
    db.close()


def test_connection_close_context_manager(tmp_path: Path, build_dir: Path) -> None:
    db_path = tmp_path / "test_connection_close_context_manager.kuzu"
    with kuzu.Database(database_path=db_path, read_only=False) as db:
        with kuzu.Connection(db) as conn:
            pass
        assert conn.is_closed
        pytest.raises(RuntimeError, conn.execute, "RETURN 1")
    assert db.is_closed


def run_long_query(conn):
    query = "UNWIND RANGE(1,1000000) AS x UNWIND RANGE(1, 1000000) AS y RETURN COUNT(x + y);"
    with pytest.raises(RuntimeError) as excinfo:
        conn.execute(query)
    assert "Interrupted" in str(excinfo.value)


def test_connection_interrupt(conn_db_readwrite: ConnDB) -> None:
    conn, _ = conn_db_readwrite
    execute_thread = threading.Thread(target=run_long_query, args=(conn,))
    execute_thread.start()
    time.sleep(5)
    conn.interrupt()
    execute_thread.join(timeout=100)
    assert not execute_thread.is_alive()



from __future__ import annotations

import hashlib
import os
import platform
import urllib.request
from pathlib import Path

import pytest
from type_aliases import ConnDB

EXTENSION_CMAKE_PREFIX = 'add_definitions(-DKUZU_EXTENSION_VERSION="'


@pytest.fixture()
def extension_extension_dir_prefix() -> str:
    system = platform.system()
    extension_extension_dir_prefix = None
    if system == "Windows":
        extension_extension_dir_prefix = "win_amd64"
    elif system == "Linux":
        extension_extension_dir_prefix = (
            "linux_arm64" if platform.machine() == "aarch64" or platform.machine() == "arm64" else "linux_amd64"
        )
    elif system == "Darwin":
        extension_extension_dir_prefix = "osx_arm64" if platform.machine() == "arm64" else "osx_amd64"
    return extension_extension_dir_prefix


def test_extension_install_httpfs(conn_db_readwrite: ConnDB, tmpdir: str, extension_extension_dir_prefix: str) -> None:
    current_dir = Path(__file__).resolve().parent
    cmake_list_file = Path(current_dir).parent.parent.parent / "CMakeLists.txt"
    extension_version = None
    with Path.open(cmake_list_file) as f:
        for line in f:
            if EXTENSION_CMAKE_PREFIX in line:
                extension_version = line.split(EXTENSION_CMAKE_PREFIX)[1].split('"')[0]
                break
    userdir = os.path.expanduser("~")  # noqa: PTH111
    extension_path = (
        Path(userdir)
        .joinpath(
            ".kuzu",
            "extension",
            extension_version,
            extension_extension_dir_prefix,
            "httpfs",
            "libhttpfs.kuzu_extension",
        )
        .resolve()
    )
    opener = urllib.request.build_opener()
    opener.addheaders = [("User-agent", "Kuzu Test Suite")]
    urllib.request.install_opener(opener)
    download_url = f"http://extension.kuzudb.com/v{extension_version}/{extension_extension_dir_prefix}/httpfs/libhttpfs.kuzu_extension"
    temp_path = Path(tmpdir) / "libhttpfs.kuzu_extension"
    urllib.request.urlretrieve(download_url, temp_path)

    conn, _ = conn_db_readwrite
    conn.execute("INSTALL hTtpFs")

    assert Path.exists(extension_path)

    extension_size = Path(extension_path).stat().st_size
    expected_size = Path(temp_path).stat().st_size

    with Path.open(extension_path, "rb") as f:
        extension_md5 = hashlib.md5(f.read()).hexdigest()
    with Path.open(temp_path, "rb") as f:
        expected_md5 = hashlib.md5(f.read()).hexdigest()

    assert extension_size == expected_size
    assert extension_md5 == expected_md5



from __future__ import annotations

import shutil
import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from test_helper import KUZU_ROOT

python_build_dir = Path(__file__).parent.parent / "build"
try:
    import kuzu
except ModuleNotFoundError:
    sys.path.append(str(python_build_dir))
    import kuzu

if TYPE_CHECKING:
    from type_aliases import ConnDB


def init_npy(conn: kuzu.Connection) -> None:
    conn.execute(
        """
        CREATE NODE TABLE npyoned (
          i64 INT64,
          i32 INT32,
          i16 INT16,
          f64 DOUBLE,
          f32 FLOAT,
          PRIMARY KEY(i64)
        );
        """
    )
    conn.execute(
        f"""
        COPY npyoned from (
          "{KUZU_ROOT}/dataset/npy-1d/one_dim_int64.npy",
          "{KUZU_ROOT}/dataset/npy-1d/one_dim_int32.npy",
          "{KUZU_ROOT}/dataset/npy-1d/one_dim_int16.npy",
          "{KUZU_ROOT}/dataset/npy-1d/one_dim_double.npy",
          "{KUZU_ROOT}/dataset/npy-1d/one_dim_float.npy") BY COLUMN;
        """
    )
    conn.execute(
        """
        CREATE NODE TABLE npytwod (
          id INT64,
          i64 INT64[3],
          i32 INT32[3],
          i16 INT16[3],
          f64 DOUBLE[3],
          f32 FLOAT[3],
          PRIMARY KEY(id)
        );
        """
    )
    conn.execute(
        f"""
        COPY npytwod FROM (
          "{KUZU_ROOT}/dataset/npy-2d/id_int64.npy",
          "{KUZU_ROOT}/dataset/npy-2d/two_dim_int64.npy",
          "{KUZU_ROOT}/dataset/npy-2d/two_dim_int32.npy",
          "{KUZU_ROOT}/dataset/npy-2d/two_dim_int16.npy",
          "{KUZU_ROOT}/dataset/npy-2d/two_dim_double.npy",
          "{KUZU_ROOT}/dataset/npy-2d/two_dim_float.npy") BY COLUMN;
        """
    )


def init_tensor(conn: kuzu.Connection) -> None:
    conn.execute(
        """
        CREATE NODE TABLE tensor (
          ID INT64,
          boolTensor BOOLEAN[],
          doubleTensor DOUBLE[][],
          intTensor INT64[][][],
          oneDimInt INT64,
          PRIMARY KEY (ID)
        );
        """
    )
    conn.execute(f'COPY tensor FROM "{KUZU_ROOT}/dataset/tensor-list/vTensor.csv" (HEADER=true)')


def init_long_str(conn: kuzu.Connection) -> None:
    conn.execute("CREATE NODE TABLE personLongString (name STRING, spouse STRING, PRIMARY KEY(name))")
    conn.execute(f'COPY personLongString FROM "{KUZU_ROOT}/dataset/long-string-pk-tests/vPerson.csv"')
    conn.execute("CREATE REL TABLE knowsLongString (FROM personLongString TO personLongString, MANY_MANY)")
    conn.execute(f'COPY knowsLongString FROM "{KUZU_ROOT}/dataset/long-string-pk-tests/eKnows.csv"')


def init_tinysnb(conn: kuzu.Connection) -> None:
    tiny_snb_path = (Path(__file__).parent / f"{KUZU_ROOT}/dataset/tinysnb").resolve()
    schema_path = tiny_snb_path / "schema.cypher"
    with schema_path.open(mode="r") as f:
        for line in f.readlines():
            line = line.strip()
            if line:
                conn.execute(line)

    copy_path = tiny_snb_path / "copy.cypher"
    with copy_path.open(mode="r") as f:
        for line in f.readlines():
            line = line.strip()
            line = line.replace("dataset/tinysnb", f"{KUZU_ROOT}/dataset/tinysnb")
            if line:
                conn.execute(line)


def init_demo(conn: kuzu.Connection) -> None:
    tiny_snb_path = (Path(__file__).parent / f"{KUZU_ROOT}/dataset/demo-db/csv").resolve()
    schema_path = tiny_snb_path / "schema.cypher"
    with schema_path.open(mode="r") as f:
        for line in f.readlines():
            line = line.strip()
            if line:
                conn.execute(line)

    copy_path = tiny_snb_path / "copy.cypher"
    with copy_path.open(mode="r") as f:
        for line in f.readlines():
            line = line.strip()
            line = line.replace("dataset/demo-db/csv", f"{KUZU_ROOT}/dataset/demo-db/csv")
            if line:
                conn.execute(line)


def init_movie_serial(conn: kuzu.Connection) -> None:
    conn.execute(
        """
        CREATE NODE TABLE moviesSerial (
          ID SERIAL,
          name STRING,
          length INT32,
          note STRING,
          PRIMARY KEY (ID)
        );"""
    )
    conn.execute(f'COPY moviesSerial from "{KUZU_ROOT}/dataset/tinysnb-serial/vMovies.csv"')


_POOL_SIZE_: int = 256 * 1024 * 1024


def init_db(path: Path) -> Path:
    if Path(path).exists():
        shutil.rmtree(path)

    conn, db = create_conn_db(path, read_only=False)
    init_tinysnb(conn)
    init_demo(conn)
    init_npy(conn)
    init_tensor(conn)
    init_long_str(conn)
    init_movie_serial(conn)
    return path


_READONLY_CONN_DB_: ConnDB | None = None
_READONLY_ASYNC_CONNECTION_: kuzu.AsyncConnection | None = None


def create_conn_db(path: Path, *, read_only: bool) -> ConnDB:
    """Return a new connection and database."""
    db = kuzu.Database(path, buffer_pool_size=_POOL_SIZE_, read_only=read_only)
    conn = kuzu.Connection(db, num_threads=4)
    return conn, db


@pytest.fixture()
def conn_db_readonly(tmp_path: Path) -> ConnDB:
    """Return a cached read-only connection and database."""
    global _READONLY_CONN_DB_
    if _READONLY_CONN_DB_ is None:
        _READONLY_CONN_DB_ = create_conn_db(init_db(tmp_path), read_only=True)
    return _READONLY_CONN_DB_


@pytest.fixture()
def conn_db_readwrite(tmp_path: Path) -> ConnDB:
    """Return a new writable connection and database."""
    return create_conn_db(init_db(tmp_path), read_only=False)


@pytest.fixture()
def async_connection_readonly(tmp_path: Path) -> kuzu.AsyncConnection:
    """Return a cached read-only async connection."""
    global _READONLY_ASYNC_CONNECTION_
    if _READONLY_ASYNC_CONNECTION_ is None:
        conn, db = create_conn_db(init_db(tmp_path), read_only=True)
        conn.close()
        _READONLY_ASYNC_CONNECTION_ = kuzu.AsyncConnection(db)
    return _READONLY_ASYNC_CONNECTION_


@pytest.fixture()
def build_dir() -> Path:
    return python_build_dir



#!/usr/bin/env python3

import os
import sys
import pandas as pd
import json


def main():
    if len(sys.argv) != 2:
        print("Usage: collect-results.py <results_dir>")
        sys.exit(1)
    if not os.path.isdir(sys.argv[1]):
        print(f"Error: {sys.argv[1]} is not a directory")
        sys.exit(1)
    results_dir = sys.argv[1]
    results_df_hash = {}
    results_exit_codes_hash = {}
    results_summary = {}
    stages = []
    for root, _, files in os.walk(results_dir):
        for csv_file in files:
            if not csv_file.endswith(".csv"):
                continue
            name_split = csv_file.split(".")
            name_split.pop()
            platform = ".".join(name_split)
            df = pd.read_csv(os.path.join(root, csv_file), header=None)
            df.columns = ["stage", "exit_code"]
            results_df_hash[platform] = df

    for platform, df in results_df_hash.items():
        for stage, exit_code in df.values:
            if stage not in stages:
                stages.append(stage)
            if platform not in results_exit_codes_hash:
                results_exit_codes_hash[platform] = {}
            results_exit_codes_hash[platform][stage] = int(exit_code)

    for platform in results_df_hash.keys():
        results_summary[platform] = []
        for stage in stages:
            status = (
                "✅"
                if stage in results_exit_codes_hash[platform]
                and results_exit_codes_hash[platform][stage] == 0
                else "❌"
            )
            results_summary[platform].append({"stage": stage, "status": status})

    summary_df = {"stage": stages}
    for platform, summary in results_summary.items():
        df = pd.DataFrame(summary)
        status = df["status"]
        summary_df[platform] = status
    summary_df = pd.DataFrame(summary_df)
    summary_df.index = summary_df["stage"]
    del summary_df["stage"]
    summary_df.index.name = None

    markdown = summary_df.to_markdown()
    with open("results.md", "w") as f:
        f.write(markdown)

    with open("results.json", "w") as f:
        json.dump(results_summary, f, indent=4)


if __name__ == "__main__":
    main()



import pyarrow as pa
import pytest
from type_aliases import ConnDB


def test_pyarrow_basic(conn_db_readonly: ConnDB) -> None:
    conn, db = conn_db_readonly
    tab = pa.Table.from_arrays(
        [
            [1, 2, 3],
            ["a", "b", "c"],
            [True, False, None],
        ],
        names=["col1", "col2", "col3"],
    )
    result = conn.execute("LOAD FROM tab RETURN * ORDER BY col1")
    assert result.get_next() == [1, "a", True]
    assert result.get_next() == [2, "b", False]
    assert result.get_next() == [3, "c", None]


def test_pyarrow_copy_from(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    tab = pa.Table.from_arrays(
        [
            pa.array([1, 2, 3], type=pa.int32()),
            pa.array(["a", "b", "c"], type=pa.string()),
            pa.array([True, False, None], type=pa.bool_()),
        ],
        names=["id", "A", "B"],
    )
    conn.execute("CREATE NODE TABLE pyarrowtab(id INT32, A STRING, B BOOL, PRIMARY KEY(id))")
    conn.execute("COPY pyarrowtab FROM tab")
    result = conn.execute("MATCH (t:pyarrowtab) RETURN t.id AS id, t.A AS A, t.B AS B ORDER BY t.id")
    assert result.get_next() == [1, "a", True]
    assert result.get_next() == [2, "b", False]
    assert result.get_next() == [3, "c", None]

    rels = pa.Table.from_arrays(
        [pa.array([1, 2, 3], type=pa.int32()), pa.array([2, 3, 1], type=pa.int32())], names=["from", "to"]
    )
    conn.execute("CREATE REL TABLE pyarrowrel(FROM pyarrowtab TO pyarrowtab)")
    conn.execute("COPY pyarrowrel FROM rels")
    result = conn.execute("MATCH (a:pyarrowtab)-[:pyarrowrel]->(b:pyarrowtab) RETURN a.id, b.id ORDER BY a.id")
    assert result.get_next() == [1, 2]
    assert result.get_next() == [2, 3]
    assert result.get_next() == [3, 1]

    conn.execute("DROP TABLE pyarrowrel")

    rels = pa.Table.from_arrays(
        [
            pa.array([1, 2, 3], type=pa.int32()),
            pa.array([2, 3, 1], type=pa.int32()),
            pa.array(["honk", "shoo", "mimimimimimimi"], type=pa.string()),
        ],
        names=["foo", "bar", "spongebobmeboy"],
    )

    conn.execute("CREATE REL TABLE pyarrowrel(FROM pyarrowtab TO pyarrowtab, a_distraction STRING)")
    conn.execute("COPY pyarrowrel FROM rels")
    result = conn.execute(
        "MATCH (a:pyarrowtab)-[r:pyarrowrel]->(b:pyarrowtab) RETURN a.id, r.a_distraction, b.id ORDER BY a.id"
    )
    assert result.get_next() == [1, "honk", 2]
    assert result.get_next() == [2, "shoo", 3]
    assert result.get_next() == [3, "mimimimimimimi", 1]


def test_pyarrow_scan_ignore_errors(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    tab = pa.Table.from_arrays(
        [
            pa.array([1, 2, 3, 1], type=pa.int32()),
        ],
        names=["id"],
    )
    conn.execute("CREATE NODE TABLE ids(id INT64, PRIMARY KEY(id))")
    conn.execute("COPY ids FROM tab(IGNORE_ERRORS=true)")

    people = conn.execute("MATCH (i:ids) RETURN i.id")
    assert people.get_next() == [1]
    assert people.get_next() == [2]
    assert people.get_next() == [3]
    assert not people.has_next()

    warnings = conn.execute("CALL show_warnings() RETURN *")
    assert warnings.get_next()[1].startswith("Found duplicated primary key value 1")
    assert not warnings.has_next()


def test_pyarrow_scan_invalid_option(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    tab = pa.Table.from_arrays(
        [
            pa.array([1, 2, 3], type=pa.int32()),
        ],
        names=["id"],
    )
    conn.execute("CREATE NODE TABLE ids(id INT64, PRIMARY KEY(id))")
    error_message = "INVALID_OPTION Option not recognized by pyArrow scanner."
    with pytest.raises(RuntimeError, match=error_message):
        conn.execute("COPY ids FROM tab(INVALID_OPTION=1)")

def test_copy_from_pyarrow_multi_pairs(conn_db_readwrite: ConnDB) -> None:
    conn, db = conn_db_readwrite
    conn.execute("CREATE NODE TABLE prof(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:prof {id: 3});")
    conn.execute("CREATE (p:prof {id: 4});")
    conn.execute("CREATE NODE TABLE student(id INT64, PRIMARY KEY(id))")
    conn.execute("CREATE (p:student {id: 2});")
    conn.execute("CREATE REL TABLE teaches(from prof to prof, from prof to student, length int64)")
    df = pa.Table.from_arrays(
        [
            pa.array([3], type=pa.int64()),
            pa.array([4], type=pa.int64()),
            pa.array([252], type=pa.int64()),
        ],
        names=["from", "to", "length"],
    )
    conn.execute("COPY teaches from df (from = 'prof', to = 'prof');")
    result = conn.execute("match (:prof)-[e:teaches]->(:prof) return e.*")
    assert result.has_next()
    assert result.get_next()[0] == 252
    assert not result.has_next()



import sys
from pathlib import Path

KUZU_ROOT = Path(__file__).parent.parent.parent.parent

if sys.platform == "win32":
    # \ in paths is not supported by kuzu's parser
    KUZU_ROOT = str(KUZU_ROOT).replace("\\", "/")
